@@ -19,9 +19,12 @@ Package bttest contains test helpers for working with the bigtable package.
 
 To use a Server, create it, and then connect to it with no security:
 (The project/instance values are ignored.)
+
 	srv, err := bttest.NewServer("localhost:0")
 	...
-	conn, err := grpc.Dial(srv.Addr, grpc.WithInsecure())
+	conn, err := grpc.Dial(
+		srv.Addr,
+		grpc.WithTransportCredentials(insecure.NewCredentials()))
 	...
 	client, err := bigtable.NewClient(ctx, proj, instance,
 	        option.WithGRPCConn(conn))
@@ -32,9 +35,11 @@ package bttest // import "cloud.google.com/go/bigtable/bttest"
 import (
 	"bytes"
 	"context"
+	"encoding/base64"
 	"encoding/binary"
 	"fmt"
 	"log"
+	"math"
 	"math/rand"
 	"net"
 	"regexp"
@@ -43,16 +48,22 @@ import (
 	"sync"
 	"time"
 
-	emptypb "github.com/golang/protobuf/ptypes/empty"
-	"github.com/golang/protobuf/ptypes/wrappers"
+	"google.golang.org/grpc/metadata"
+	"google.golang.org/protobuf/proto"
+
+	longrunning "cloud.google.com/go/longrunning/autogen/longrunningpb"
 	"github.com/google/btree"
 	btapb "google.golang.org/genproto/googleapis/bigtable/admin/v2"
 	btpb "google.golang.org/genproto/googleapis/bigtable/v2"
-	"google.golang.org/genproto/googleapis/longrunning"
 	statpb "google.golang.org/genproto/googleapis/rpc/status"
 	"google.golang.org/grpc"
 	"google.golang.org/grpc/codes"
 	"google.golang.org/grpc/status"
+	"google.golang.org/protobuf/types/known/anypb"
+	"google.golang.org/protobuf/types/known/durationpb"
+	"google.golang.org/protobuf/types/known/emptypb"
+	"google.golang.org/protobuf/types/known/timestamppb"
+	"google.golang.org/protobuf/types/known/wrapperspb"
 	"rsc.io/binaryregexp"
 )
 
@@ -61,12 +72,12 @@ const (
 	minValidMilliSeconds = 0
 
 	// MilliSeconds field of the max valid Timestamp.
-	maxValidMilliSeconds = int64(time.Millisecond) * 253402300800
+	// Must match the max value of type TimestampMicros (int64)
+	// truncated to the millis granularity by subtracting a remainder of 1000.
+	maxValidMilliSeconds = math.MaxInt64 - math.MaxInt64%1000
 )
 
-var (
-	validLabelTransformer = regexp.MustCompile(`[a-z0-9\-]{1,15}`)
-)
+var validLabelTransformer = regexp.MustCompile(`[a-z0-9\-]{1,15}`)
 
 // Server is an in-memory Cloud Bigtable fake.
 // It is unauthenticated, and only a rough approximation.
@@ -144,9 +155,10 @@ func (s *server) CreateTable(ctx context.Context, req *btapb.CreateTableRequest)
 	s.mu.Unlock()
 
 	ct := &btapb.Table{
-		Name:           tbl,
-		ColumnFamilies: req.GetTable().GetColumnFamilies(),
-		Granularity:    req.GetTable().GetGranularity(),
+		Name:               tbl,
+		ColumnFamilies:     req.GetTable().GetColumnFamilies(),
+		Granularity:        req.GetTable().GetGranularity(),
+		DeletionProtection: req.GetTable().GetDeletionProtection(),
 	}
 	if ct.Granularity == 0 {
 		ct.Granularity = btapb.Table_MILLIS
@@ -184,8 +196,9 @@ func (s *server) GetTable(ctx context.Context, req *btapb.GetTableRequest) (*bta
 	}
 
 	return &btapb.Table{
-		Name:           tbl,
-		ColumnFamilies: toColumnFamilies(tblIns.columnFamilies()),
+		Name:               tbl,
+		ColumnFamilies:     toColumnFamilies(tblIns.columnFamilies()),
+		DeletionProtection: tblIns.isProtected,
 	}, nil
 }
 
@@ -195,10 +208,73 @@ func (s *server) DeleteTable(ctx context.Context, req *btapb.DeleteTableRequest)
 	if _, ok := s.tables[req.Name]; !ok {
 		return nil, status.Errorf(codes.NotFound, "table %q not found", req.Name)
 	}
+	if s.tables[req.Name].isProtected {
+		return nil, status.Errorf(codes.FailedPrecondition, "table %q is protected from deletion", req.Name)
+	}
 	delete(s.tables, req.Name)
 	return &emptypb.Empty{}, nil
 }
 
+func (s *server) UpdateTable(ctx context.Context, req *btapb.UpdateTableRequest) (*longrunning.Operation, error) {
+	s.mu.Lock()
+	defer s.mu.Unlock()
+
+	updateMask := req.UpdateMask
+	if updateMask == nil {
+		return nil, status.Errorf(codes.InvalidArgument, "UpdateTableRequest.UpdateMask required for table update")
+	}
+
+	var utr *btapb.Table
+	if !updateMask.IsValid(utr) {
+		return nil, status.Errorf(codes.InvalidArgument, "incorrect path in UpdateMask; got: %v\n", updateMask)
+	}
+
+	tbl, ok := s.tables[req.GetTable().GetName()]
+	if !ok {
+		return nil, status.Errorf(codes.NotFound, "table %q not found", req.GetTable().GetName())
+	}
+	tbl.mu.Lock()
+	defer tbl.mu.Unlock()
+
+	tbl.isProtected = req.GetTable().GetDeletionProtection()
+
+	res := &longrunning.Operation_Response{}
+	lro := &longrunning.Operation{
+		Name:   "projects/my-project/operations/1234",
+		Done:   true,
+		Result: res,
+	}
+
+	a, err := anypb.New(&btapb.UpdateTableMetadata{
+		Name:      req.GetTable().GetName(),
+		StartTime: timestamppb.Now(),
+		EndTime:   timestamppb.Now(),
+	})
+	if err != nil {
+		lro.Result = &longrunning.Operation_Error{
+			Error: &statpb.Status{
+				Code:    int32(codes.Internal),
+				Message: err.Error(),
+			},
+		}
+		return lro, nil
+	}
+	anyTable, err := anypb.New(req.GetTable())
+	if err != nil {
+		lro.Result = &longrunning.Operation_Error{
+			Error: &statpb.Status{
+				Code:    int32(codes.Internal),
+				Message: err.Error(),
+			},
+		}
+		return lro, nil
+	}
+
+	lro.Metadata = a
+	res.Response = anyTable
+	return lro, nil
+}
+
 func (s *server) ModifyColumnFamilies(ctx context.Context, req *btapb.ModifyColumnFamiliesRequest) (*btapb.Table, error) {
 	s.mu.Lock()
 	tbl, ok := s.tables[req.Name]
@@ -210,16 +286,22 @@ func (s *server) ModifyColumnFamilies(ctx context.Context, req *btapb.ModifyColu
 	tbl.mu.Lock()
 	defer tbl.mu.Unlock()
 
+	// Check table protection status
+	if tbl.isProtected {
+		for _, mod := range req.Modifications {
+			// Cannot delete columns from a protected table
+			if mod.GetDrop() {
+				return nil, status.Errorf(codes.FailedPrecondition, "table %q is protected from deletion", req.Name)
+			}
+		}
+	}
+
 	for _, mod := range req.Modifications {
 		if create := mod.GetCreate(); create != nil {
 			if _, ok := tbl.families[mod.Id]; ok {
 				return nil, status.Errorf(codes.AlreadyExists, "family %q already exists", mod.Id)
 			}
-			newcf := &columnFamily{
-				name:   req.Name + "/columnFamilies/" + mod.Id,
-				order:  tbl.counter,
-				gcRule: create.GcRule,
-			}
+			newcf := newColumnFamily(req.Name+"/columnFamilies/"+mod.Id, tbl.counter, create)
 			tbl.counter++
 			tbl.families[mod.Id] = newcf
 		} else if mod.GetDrop() {
@@ -227,14 +309,20 @@ func (s *server) ModifyColumnFamilies(ctx context.Context, req *btapb.ModifyColu
 				return nil, fmt.Errorf("can't delete unknown family %q", mod.Id)
 			}
 			delete(tbl.families, mod.Id)
+
+			// Purge all data for this column family
+			tbl.rows.Ascend(func(i btree.Item) bool {
+				r := i.(*row)
+				r.mu.Lock()
+				defer r.mu.Unlock()
+				delete(r.families, mod.Id)
+				return true
+			})
 		} else if modify := mod.GetUpdate(); modify != nil {
 			if _, ok := tbl.families[mod.Id]; !ok {
 				return nil, fmt.Errorf("no such family %q", mod.Id)
 			}
-			newcf := &columnFamily{
-				name:   req.Name + "/columnFamilies/" + mod.Id,
-				gcRule: modify.GcRule,
-			}
+			newcf := newColumnFamily(req.Name+"/columnFamilies/"+mod.Id, 0, modify)
 			// assume that we ALWAYS want to replace by the new setting
 			// we may need partial update through
 			tbl.families[mod.Id] = newcf
@@ -251,12 +339,14 @@ func (s *server) ModifyColumnFamilies(ctx context.Context, req *btapb.ModifyColu
 
 func (s *server) DropRowRange(ctx context.Context, req *btapb.DropRowRangeRequest) (*emptypb.Empty, error) {
 	s.mu.Lock()
-	defer s.mu.Unlock()
 	tbl, ok := s.tables[req.Name]
+	s.mu.Unlock()
 	if !ok {
 		return nil, status.Errorf(codes.NotFound, "table %q not found", req.Name)
 	}
 
+	tbl.mu.Lock()
+	defer tbl.mu.Unlock()
 	if req.GetDeleteAllDataFromTable() {
 		tbl.rows = btree.New(btreeDegree)
 	} else {
@@ -323,14 +413,41 @@ func (s *server) SnapshotTable(context.Context, *btapb.SnapshotTableRequest) (*l
 func (s *server) GetSnapshot(context.Context, *btapb.GetSnapshotRequest) (*btapb.Snapshot, error) {
 	return nil, status.Errorf(codes.Unimplemented, "the emulator does not currently support snapshots")
 }
+
 func (s *server) ListSnapshots(context.Context, *btapb.ListSnapshotsRequest) (*btapb.ListSnapshotsResponse, error) {
 	return nil, status.Errorf(codes.Unimplemented, "the emulator does not currently support snapshots")
 }
+
 func (s *server) DeleteSnapshot(context.Context, *btapb.DeleteSnapshotRequest) (*emptypb.Empty, error) {
 	return nil, status.Errorf(codes.Unimplemented, "the emulator does not currently support snapshots")
 }
 
+func featureFlagsFromContext(context context.Context) *btpb.FeatureFlags {
+	ff := &btpb.FeatureFlags{}
+
+	var md, ok = metadata.FromIncomingContext(context)
+	if !ok {
+		return ff
+	}
+
+	features := md.Get("bigtable-features")
+	if len(features) == 0 {
+		return ff
+	}
+
+	dec, err := base64.URLEncoding.DecodeString(features[0])
+	if err != nil {
+		return ff
+	}
+
+	_ = proto.Unmarshal(dec, ff)
+	return ff
+}
+
 func (s *server) ReadRows(req *btpb.ReadRowsRequest, stream btpb.Bigtable_ReadRowsServer) error {
+	featureFlags := featureFlagsFromContext(stream.Context())
+
+	start := time.Now()
 	s.mu.Lock()
 	tbl, ok := s.tables[req.TableName]
 	s.mu.Unlock()
@@ -405,41 +522,86 @@ func (s *server) ReadRows(req *btpb.ReadRowsRequest, stream btpb.Bigtable_ReadRo
 			rows = append(rows, r)
 		}
 	}
-	sort.Sort(byRowKey(rows))
+	if req.Reversed {
+		if !featureFlags.ReverseScans {
+			return status.Errorf(codes.Unimplemented, "Client doesn't support reverse scans yet")
+		}
+
+		sort.Sort(sort.Reverse(byRowKey(rows)))
+	} else {
+		sort.Sort(byRowKey(rows))
+	}
 
 	limit := int(req.RowsLimit)
-	count := 0
+	if limit == 0 {
+		limit = len(rows)
+	}
+
+	iterStats := &btpb.ReadIterationStats{}
+
 	for _, r := range rows {
-		if limit > 0 && count >= limit {
-			return nil
+		if int(iterStats.RowsReturnedCount) >= limit {
+			break
 		}
-		streamed, err := streamRow(stream, r, req.Filter)
-		if err != nil {
+
+		if err := streamRow(stream, r, req.Filter, iterStats, featureFlags); err != nil {
 			return err
 		}
-		if streamed {
-			count++
+	}
+
+	elapsed := time.Since(start)
+	if req.RequestStatsView == btpb.ReadRowsRequest_REQUEST_STATS_FULL {
+		rrr := &btpb.ReadRowsResponse{}
+		rrr.RequestStats = &btpb.RequestStats{
+			StatsView: &btpb.RequestStats_FullReadStatsView{
+				FullReadStatsView: &btpb.FullReadStatsView{
+					ReadIterationStats: iterStats,
+					RequestLatencyStats: &btpb.RequestLatencyStats{
+						FrontendServerLatency: durationpb.New(elapsed),
+					},
+				},
+			},
 		}
+
+		return stream.Send(rrr)
 	}
 	return nil
 }
 
 // streamRow filters the given row and sends it via the given stream.
 // Returns true if at least one cell matched the filter and was streamed, false otherwise.
-func streamRow(stream btpb.Bigtable_ReadRowsServer, r *row, f *btpb.RowFilter) (bool, error) {
+func streamRow(stream btpb.Bigtable_ReadRowsServer, r *row, f *btpb.RowFilter, s *btpb.ReadIterationStats, ff *btpb.FeatureFlags) error {
 	r.mu.Lock()
 	nr := r.copy()
 	r.mu.Unlock()
 	r = nr
 
+	s.RowsSeenCount++
+	// Count cells in the row before filtering for CellsSeenCount.
+	for _, fam := range r.families {
+		for _, colName := range fam.colNames {
+			s.CellsSeenCount += int64(len(fam.cells[colName]))
+		}
+	}
+
 	match, err := filterRow(f, r)
 	if err != nil {
-		return false, err
+		return err
 	}
 	if !match {
-		return false, nil
+		// if the client requested it, send last_scanned_row responses for rows that didn't match a filter
+		if ff.LastScannedRowResponses {
+			rrr := &btpb.ReadRowsResponse{
+				LastScannedRowKey: []byte(r.key),
+			}
+			return stream.Send(rrr)
+		}
+
+		return nil
 	}
 
+	s.RowsReturnedCount++
+
 	rrr := &btpb.ReadRowsResponse{}
 	families := r.sortedFamilies()
 	for _, fam := range families {
@@ -448,11 +610,12 @@ func streamRow(stream btpb.Bigtable_ReadRowsServer, r *row, f *btpb.RowFilter) (
 			if len(cells) == 0 {
 				continue
 			}
+			s.CellsReturnedCount += int64(len(cells))
 			for _, cell := range cells {
 				rrr.Chunks = append(rrr.Chunks, &btpb.ReadRowsResponse_CellChunk{
 					RowKey:          []byte(r.key),
-					FamilyName:      &wrappers.StringValue{Value: fam.name},
-					Qualifier:       &wrappers.BytesValue{Value: []byte(colName)},
+					FamilyName:      &wrapperspb.StringValue{Value: fam.name},
+					Qualifier:       &wrapperspb.BytesValue{Value: []byte(colName)},
 					TimestampMicros: cell.ts,
 					Value:           cell.value,
 					Labels:          cell.labels,
@@ -466,7 +629,7 @@ func streamRow(stream btpb.Bigtable_ReadRowsServer, r *row, f *btpb.RowFilter) (
 		rrr.Chunks[len(rrr.Chunks)-1].RowStatus = &btpb.ReadRowsResponse_CellChunk_CommitRow{CommitRow: true}
 	}
 
-	return true, stream.Send(rrr)
+	return stream.Send(rrr)
 }
 
 // filterRow modifies a row with the given filter. Returns true if at least one cell from the row matches,
@@ -478,10 +641,19 @@ func filterRow(f *btpb.RowFilter, r *row) (bool, error) {
 	// Handle filters that apply beyond just including/excluding cells.
 	switch f := f.Filter.(type) {
 	case *btpb.RowFilter_BlockAllFilter:
-		return !f.BlockAllFilter, nil
+		if !f.BlockAllFilter {
+			return false, status.Errorf(codes.InvalidArgument, "block_all_filter must be true if set")
+		}
+		return false, nil
 	case *btpb.RowFilter_PassAllFilter:
-		return f.PassAllFilter, nil
+		if !f.PassAllFilter {
+			return false, status.Errorf(codes.InvalidArgument, "pass_all_filter must be true if set")
+		}
+		return true, nil
 	case *btpb.RowFilter_Chain_:
+		if len(f.Chain.Filters) < 2 {
+			return false, status.Errorf(codes.InvalidArgument, "Chain must contain at least two RowFilters")
+		}
 		for _, sub := range f.Chain.Filters {
 			match, err := filterRow(sub, r)
 			if err != nil {
@@ -493,6 +665,9 @@ func filterRow(f *btpb.RowFilter, r *row) (bool, error) {
 		}
 		return true, nil
 	case *btpb.RowFilter_Interleave_:
+		if len(f.Interleave.Filters) < 2 {
+			return false, status.Errorf(codes.InvalidArgument, "Interleave must contain at least two RowFilters")
+		}
 		srs := make([]*row, 0, len(f.Interleave.Filters))
 		for _, sub := range f.Interleave.Filters {
 			sr := r.copy()
@@ -525,6 +700,9 @@ func filterRow(f *btpb.RowFilter, r *row) (bool, error) {
 		return count > 0, nil
 	case *btpb.RowFilter_CellsPerColumnLimitFilter:
 		lim := int(f.CellsPerColumnLimitFilter)
+		if lim <= 0 {
+			return false, status.Errorf(codes.InvalidArgument, "Error in field 'cells_per_column_limit_filter' : argument must be > 0")
+		}
 		for _, fam := range r.families {
 			for col, cs := range fam.cells {
 				if len(cs) > lim {
@@ -549,9 +727,12 @@ func filterRow(f *btpb.RowFilter, r *row) (bool, error) {
 		}
 		return filterRow(f.Condition.FalseFilter, r)
 	case *btpb.RowFilter_RowKeyRegexFilter:
+		if len(f.RowKeyRegexFilter) == 0 {
+			return false, status.Errorf(codes.InvalidArgument, "Error in field 'row_key_regex_filter' : argument must not be empty")
+		}
 		rx, err := newRegexp(f.RowKeyRegexFilter)
 		if err != nil {
-			return false, status.Errorf(codes.InvalidArgument, "Error in field 'rowkey_regex_filter' : %v", err)
+			return false, status.Errorf(codes.InvalidArgument, "Error in field 'row_key_regex_filter' : %v", err)
 		}
 		if !rx.MatchString(r.key) {
 			return false, nil
@@ -559,6 +740,9 @@ func filterRow(f *btpb.RowFilter, r *row) (bool, error) {
 	case *btpb.RowFilter_CellsPerRowLimitFilter:
 		// Grab the first n cells in the row.
 		lim := int(f.CellsPerRowLimitFilter)
+		if lim <= 0 {
+			return false, status.Errorf(codes.InvalidArgument, "Error in field 'cells_per_row_limit_filter' : argument must be > 0")
+		}
 		for _, fam := range r.families {
 			for _, col := range fam.colNames {
 				cs := fam.cells[col]
@@ -586,7 +770,10 @@ func filterRow(f *btpb.RowFilter, r *row) (bool, error) {
 				offset -= len(cs)
 			}
 		}
-		return true, nil
+		// If we get here, we have to have consumed all of the cells,
+		// otherwise, we would have returned above.  We're not generating
+		// a row, so false.
+		return false, nil
 	case *btpb.RowFilter_RowSampleFilter:
 		// The row sample filter "matches all cells from a row with probability
 		// p, and matches no cells from the row with probability 1-p."
@@ -715,10 +902,6 @@ func includeCell(f *btpb.RowFilter, fam, col string, cell cell) (bool, error) {
 		}
 		return inRangeStart() && inRangeEnd(), nil
 	case *btpb.RowFilter_TimestampRangeFilter:
-		// Server should only support millisecond precision.
-		if f.TimestampRangeFilter.StartTimestampMicros%int64(time.Millisecond/time.Microsecond) != 0 || f.TimestampRangeFilter.EndTimestampMicros%int64(time.Millisecond/time.Microsecond) != 0 {
-			return false, status.Errorf(codes.InvalidArgument, "Error in field 'timestamp_range_filter'. Maximum precision allowed in filter is millisecond.\nGot:\nStart: %v\nEnd: %v", f.TimestampRangeFilter.StartTimestampMicros, f.TimestampRangeFilter.EndTimestampMicros)
-		}
 		// Lower bound is inclusive and defaults to 0, upper bound is exclusive and defaults to infinity.
 		return cell.ts >= f.TimestampRangeFilter.StartTimestampMicros &&
 			(f.TimestampRangeFilter.EndTimestampMicros == 0 || cell.ts < f.TimestampRangeFilter.EndTimestampMicros), nil
@@ -744,8 +927,35 @@ func includeCell(f *btpb.RowFilter, fam, col string, cell cell) (bool, error) {
 	}
 }
 
+// escapeUTF is used to escape non-ASCII characters in pattern strings passed
+// to binaryregexp. This makes regexp column and row key matching work more
+// closely to what's seen with the real BigTable.
+func escapeUTF(in []byte) []byte {
+	var toEsc int
+	for _, c := range in {
+		if c > 127 {
+			toEsc++
+		}
+	}
+	if toEsc == 0 {
+		return in
+	}
+	// Each escaped byte becomes 4 bytes (byte a1 becomes \xA1)
+	out := make([]byte, 0, len(in)+3*toEsc)
+	for _, c := range in {
+		if c > 127 {
+			h, l := c>>4, c&0xF
+			const conv = "0123456789ABCDEF"
+			out = append(out, '\\', 'x', conv[h], conv[l])
+		} else {
+			out = append(out, c)
+		}
+	}
+	return out
+}
+
 func newRegexp(pat []byte) (*binaryregexp.Regexp, error) {
-	re, err := binaryregexp.Compile("^(?:" + string(pat) + ")$") // match entire target
+	re, err := binaryregexp.Compile("^(?:" + string(escapeUTF(pat)) + ")$") // match entire target
 	if err != nil {
 		log.Printf("Bad pattern %q: %v", pat, err)
 	}
@@ -753,6 +963,12 @@ func newRegexp(pat []byte) (*binaryregexp.Regexp, error) {
 }
 
 func (s *server) MutateRow(ctx context.Context, req *btpb.MutateRowRequest) (*btpb.MutateRowResponse, error) {
+	if len(req.Mutations) == 0 {
+		return nil, status.Errorf(
+			codes.InvalidArgument,
+			"No mutations provided",
+		)
+	}
 	s.mu.Lock()
 	tbl, ok := s.tables[req.TableName]
 	s.mu.Unlock()
@@ -770,6 +986,16 @@ func (s *server) MutateRow(ctx context.Context, req *btpb.MutateRowRequest) (*bt
 }
 
 func (s *server) MutateRows(req *btpb.MutateRowsRequest, stream btpb.Bigtable_MutateRowsServer) error {
+	nMutations := 0
+	for _, entry := range req.Entries {
+		nMutations += len(entry.Mutations)
+	}
+	if nMutations == 0 {
+		return status.Errorf(
+			codes.InvalidArgument,
+			"No mutations provided",
+		)
+	}
 	s.mu.Lock()
 	tbl, ok := s.tables[req.TableName]
 	s.mu.Unlock()
@@ -840,17 +1066,29 @@ func (s *server) CheckAndMutateRow(ctx context.Context, req *btpb.CheckAndMutate
 	return res, nil
 }
 
+func (s *server) PingAndWarm(ctx context.Context, req *btpb.PingAndWarmRequest) (*btpb.PingAndWarmResponse, error) {
+	return &btpb.PingAndWarmResponse{}, nil
+}
+
 // applyMutations applies a sequence of mutations to a row.
 // fam should be a snapshot of the keys of tbl.families.
 // It assumes r.mu is locked.
 func applyMutations(tbl *table, r *row, muts []*btpb.Mutation, fs map[string]*columnFamily) error {
+	if len(r.key) == 0 {
+		return status.Errorf(
+			codes.InvalidArgument,
+			"Row keys must be non-empty",
+		)
+	}
+
 	for _, mut := range muts {
 		switch mut := mut.Mutation.(type) {
 		default:
 			return fmt.Errorf("can't handle mutation type %T", mut)
 		case *btpb.Mutation_SetCell_:
 			set := mut.SetCell
-			if _, ok := fs[set.FamilyName]; !ok {
+			var cf, ok = fs[set.FamilyName]
+			if !ok {
 				return fmt.Errorf("unknown family %q", set.FamilyName)
 			}
 			ts := set.TimestampMicros
@@ -865,7 +1103,36 @@ func applyMutations(tbl *table, r *row, muts []*btpb.Mutation, fs map[string]*co
 
 			newCell := cell{ts: ts, value: set.Value}
 			f := r.getOrCreateFamily(fam, fs[fam].order)
-			f.cells[col] = appendOrReplaceCell(f.cellsByColumn(col), newCell)
+			f.cells[col] = appendOrReplaceCell(f.cellsByColumn(col), newCell, cf)
+		case *btpb.Mutation_AddToCell_:
+			add := mut.AddToCell
+			var cf, ok = fs[add.FamilyName]
+			if !ok {
+				return fmt.Errorf("unknown family %q", add.FamilyName)
+			}
+			if cf.valueType == nil || cf.valueType.GetAggregateType() == nil {
+				return fmt.Errorf("illegal attempt to use AddToCell on non-aggregate cell")
+			}
+			ts := add.Timestamp.GetRawTimestampMicros()
+			if ts < 0 {
+				return fmt.Errorf("AddToCell must set timestamp >= 0")
+			}
+
+			fam := add.FamilyName
+			col := string(add.GetColumnQualifier().GetRawValue())
+
+			var value []byte
+			switch v := add.Input.Kind.(type) {
+			case *btpb.Value_IntValue:
+				value = binary.BigEndian.AppendUint64(value, uint64(v.IntValue))
+			default:
+				return fmt.Errorf("only int64 values are supported")
+			}
+
+			newCell := cell{ts: ts, value: value}
+			f := r.getOrCreateFamily(fam, fs[fam].order)
+			f.cells[col] = appendOrReplaceCell(f.cellsByColumn(col), newCell, cf)
+
 		case *btpb.Mutation_DeleteFromColumn_:
 			del := mut.DeleteFromColumn
 			if _, ok := fs[del.FamilyName]; !ok {
@@ -941,16 +1208,18 @@ func newTimestamp() int64 {
 	return ts
 }
 
-func appendOrReplaceCell(cs []cell, newCell cell) []cell {
+func appendOrReplaceCell(cs []cell, newCell cell, cf *columnFamily) []cell {
 	replaced := false
 	for i, cell := range cs {
 		if cell.ts == newCell.ts {
+			newCell.value = cf.updateFn(cs[i].value, newCell.value)
 			cs[i] = newCell
 			replaced = true
 			break
 		}
 	}
 	if !replaced {
+		newCell.value = cf.initFn(newCell.value)
 		cs = append(cs, newCell)
 	}
 	sort.Sort(byDescTS(cs))
@@ -977,7 +1246,8 @@ func (s *server) ReadModifyWriteRow(ctx context.Context, req *btpb.ReadModifyWri
 	// Assume all mutations apply to the most recent version of the cell.
 	// TODO(dsymonds): Verify this assumption and document it in the proto.
 	for _, rule := range req.Rules {
-		if _, ok := fs[rule.FamilyName]; !ok {
+		var cf, ok = fs[rule.FamilyName]
+		if !ok {
 			return nil, fmt.Errorf("unknown family %q", rule.FamilyName)
 		}
 
@@ -1020,7 +1290,7 @@ func (s *server) ReadModifyWriteRow(ctx context.Context, req *btpb.ReadModifyWri
 		}
 
 		// Store the new cell
-		f.cells[col] = appendOrReplaceCell(f.cellsByColumn(col), newCell)
+		f.cells[col] = appendOrReplaceCell(f.cellsByColumn(col), newCell, cf)
 
 		// Store a copy for the result row
 		resultFamily := resultRow.getOrCreateFamily(fam, fs[fam].order)
@@ -1071,6 +1341,8 @@ func (s *server) SampleRowKeys(req *btpb.SampleRowKeysRequest, stream btpb.Bigta
 	i := 0
 	tbl.rows.Ascend(func(it btree.Item) bool {
 		row := it.(*row)
+		row.mu.Lock()
+		defer row.mu.Unlock()
 		if i == tbl.rows.Len()-1 || rand.Int31n(100) == 0 {
 			resp := &btpb.SampleRowKeysResponse{
 				RowKey:      []byte(row.key),
@@ -1127,10 +1399,11 @@ func (s *server) gcloop(done <-chan int) {
 }
 
 type table struct {
-	mu       sync.RWMutex
-	counter  uint64                   // increment by 1 when a new family is created
-	families map[string]*columnFamily // keyed by plain family name
-	rows     *btree.BTree             // indexed by row key
+	mu          sync.RWMutex
+	counter     uint64                   // increment by 1 when a new family is created
+	families    map[string]*columnFamily // keyed by plain family name
+	rows        *btree.BTree             // indexed by row key
+	isProtected bool                     // whether this table has deletion protection
 }
 
 const btreeDegree = 16
@@ -1140,18 +1413,15 @@ func newTable(ctr *btapb.CreateTableRequest) *table {
 	c := uint64(0)
 	if ctr.Table != nil {
 		for id, cf := range ctr.Table.ColumnFamilies {
-			fams[id] = &columnFamily{
-				name:   ctr.Parent + "/columnFamilies/" + id,
-				order:  c,
-				gcRule: cf.GcRule,
-			}
+			fams[id] = newColumnFamily(ctr.Parent+"/columnFamilies/"+id, c, cf)
 			c++
 		}
 	}
 	return &table{
-		families: fams,
-		counter:  c,
-		rows:     btree.New(btreeDegree),
+		families:    fams,
+		counter:     c,
+		rows:        btree.New(btreeDegree),
+		isProtected: ctr.GetTable().GetDeletionProtection(),
 	}
 }
 
@@ -1197,6 +1467,27 @@ func (t *table) mutableRow(key string) *row {
 }
 
 func (t *table) gc() {
+	toDelete := t.gcReadOnly()
+	if len(toDelete) == 0 {
+		return
+	}
+
+	// We delete rows that no longer have any cells
+	t.mu.Lock()
+	defer t.mu.Unlock()
+	for _, i := range toDelete {
+		r := i.(*row)
+		// Make sure the row still has no cells. We've not been holding a lock
+		// so it could have changed since we checked it.
+		r.mu.Lock()
+		if len(r.families) == 0 {
+			t.rows.Delete(i)
+		}
+		r.mu.Unlock()
+	}
+}
+
+func (t *table) gcReadOnly() (toDelete []btree.Item) {
 	// This method doesn't add or remove rows, so we only need a read lock for the table.
 	t.mu.RLock()
 	defer t.mu.RUnlock()
@@ -1209,16 +1500,23 @@ func (t *table) gc() {
 		}
 	}
 	if len(rules) == 0 {
-		return
+		return nil
 	}
 
+	// It isn't clear whether it's safe to delete within the iterator, so we do
+	// not
 	t.rows.Ascend(func(i btree.Item) bool {
 		r := i.(*row)
 		r.mu.Lock()
 		r.gc(rules)
+		if len(r.families) == 0 {
+			toDelete = append(toDelete, i)
+		}
 		r.mu.Unlock()
 		return true
 	})
+
+	return toDelete
 }
 
 type byRowKey []*row
@@ -1304,7 +1602,15 @@ func (r *row) gc(rules map[string]*btapb.GcRule) {
 			continue
 		}
 		for col, cs := range fam.cells {
-			r.families[fam.name].cells[col] = applyGC(cs, rule)
+			cs = applyGC(cs, rule)
+			if len(cs) == 0 {
+				delete(fam.cells, col)
+			} else {
+				fam.cells[col] = cs
+			}
+		}
+		if len(fam.cells) == 0 {
+			delete(r.families, fam.name)
 		}
 	}
 }
@@ -1406,15 +1712,49 @@ func (b byDescTS) Len() int           { return len(b) }
 func (b byDescTS) Swap(i, j int)      { b[i], b[j] = b[j], b[i] }
 func (b byDescTS) Less(i, j int) bool { return b[i].ts > b[j].ts }
 
+func newColumnFamily(name string, order uint64, cf *btapb.ColumnFamily) *columnFamily {
+	var updateFn = func(_, newVal []byte) []byte {
+		return newVal
+	}
+	if cf.ValueType != nil {
+		switch v := cf.ValueType.Kind.(type) {
+		case *btapb.Type_AggregateType:
+			switch v.AggregateType.Aggregator.(type) {
+			case *btapb.Type_Aggregate_Sum_:
+				updateFn = func(existing, newVal []byte) []byte {
+					existingInt := int64(binary.BigEndian.Uint64(existing))
+					newInt := int64(binary.BigEndian.Uint64(newVal))
+					return binary.BigEndian.AppendUint64([]byte{}, uint64(existingInt+newInt))
+				}
+			}
+		default:
+		}
+	}
+	return &columnFamily{
+		name:      name,
+		order:     order,
+		gcRule:    cf.GcRule,
+		valueType: cf.ValueType,
+		updateFn:  updateFn,
+		initFn: func(newVal []byte) []byte {
+			return newVal
+		},
+	}
+}
+
 type columnFamily struct {
-	name   string
-	order  uint64 // Creation order of column family
-	gcRule *btapb.GcRule
+	name      string
+	order     uint64 // Creation order of column family
+	gcRule    *btapb.GcRule
+	valueType *btapb.Type
+	updateFn  func(existing, newVal []byte) []byte
+	initFn    func(newVal []byte) []byte
 }
 
 func (c *columnFamily) proto() *btapb.ColumnFamily {
 	return &btapb.ColumnFamily{
-		GcRule: c.gcRule,
+		GcRule:    c.gcRule,
+		ValueType: c.valueType,
 	}
 }
 
