@@ -16,24 +16,27 @@ package firestore
 
 import (
 	"context"
+	"fmt"
 	"math"
 	"sort"
 	"testing"
 
+	pb "cloud.google.com/go/firestore/apiv1/firestorepb"
 	"cloud.google.com/go/internal/pretty"
-	tspb "github.com/golang/protobuf/ptypes/timestamp"
-	"github.com/golang/protobuf/ptypes/wrappers"
-	pb "google.golang.org/genproto/googleapis/firestore/v1"
+	"github.com/google/go-cmp/cmp"
+	"google.golang.org/protobuf/testing/protocmp"
+	tspb "google.golang.org/protobuf/types/known/timestamppb"
+	"google.golang.org/protobuf/types/known/wrapperspb"
 )
 
 func TestFilterToProto(t *testing.T) {
 	for _, test := range []struct {
-		in   filter
+		in   EntityFilter
 		want *pb.StructuredQuery_Filter
 	}{
 		{
-			filter{[]string{"a"}, ">", 1},
-			&pb.StructuredQuery_Filter{FilterType: &pb.StructuredQuery_Filter_FieldFilter{
+			in: PropertyFilter{"a", ">", 1},
+			want: &pb.StructuredQuery_Filter{FilterType: &pb.StructuredQuery_Filter_FieldFilter{
 				FieldFilter: &pb.StructuredQuery_FieldFilter{
 					Field: &pb.StructuredQuery_FieldReference{FieldPath: "a"},
 					Op:    pb.StructuredQuery_FieldFilter_GREATER_THAN,
@@ -42,8 +45,8 @@ func TestFilterToProto(t *testing.T) {
 			}},
 		},
 		{
-			filter{[]string{"a"}, "==", nil},
-			&pb.StructuredQuery_Filter{FilterType: &pb.StructuredQuery_Filter_UnaryFilter{
+			in: PropertyFilter{"a", "==", nil},
+			want: &pb.StructuredQuery_Filter{FilterType: &pb.StructuredQuery_Filter_UnaryFilter{
 				UnaryFilter: &pb.StructuredQuery_UnaryFilter{
 					OperandType: &pb.StructuredQuery_UnaryFilter_Field{
 						Field: &pb.StructuredQuery_FieldReference{FieldPath: "a"},
@@ -53,8 +56,8 @@ func TestFilterToProto(t *testing.T) {
 			}},
 		},
 		{
-			filter{[]string{"a"}, "==", math.NaN()},
-			&pb.StructuredQuery_Filter{FilterType: &pb.StructuredQuery_Filter_UnaryFilter{
+			in: PropertyFilter{"a", "==", math.NaN()},
+			want: &pb.StructuredQuery_Filter{FilterType: &pb.StructuredQuery_Filter_UnaryFilter{
 				UnaryFilter: &pb.StructuredQuery_UnaryFilter{
 					OperandType: &pb.StructuredQuery_UnaryFilter_Field{
 						Field: &pb.StructuredQuery_FieldReference{FieldPath: "a"},
@@ -63,6 +66,166 @@ func TestFilterToProto(t *testing.T) {
 				},
 			}},
 		},
+		{
+			in: PropertyPathFilter{[]string{"a"}, ">", 1},
+			want: &pb.StructuredQuery_Filter{FilterType: &pb.StructuredQuery_Filter_FieldFilter{
+				FieldFilter: &pb.StructuredQuery_FieldFilter{
+					Field: &pb.StructuredQuery_FieldReference{FieldPath: "a"},
+					Op:    pb.StructuredQuery_FieldFilter_GREATER_THAN,
+					Value: intval(1),
+				},
+			}},
+		},
+		{
+			in: PropertyPathFilter{[]string{"a"}, "==", nil},
+			want: &pb.StructuredQuery_Filter{FilterType: &pb.StructuredQuery_Filter_UnaryFilter{
+				UnaryFilter: &pb.StructuredQuery_UnaryFilter{
+					OperandType: &pb.StructuredQuery_UnaryFilter_Field{
+						Field: &pb.StructuredQuery_FieldReference{FieldPath: "a"},
+					},
+					Op: pb.StructuredQuery_UnaryFilter_IS_NULL,
+				},
+			}},
+		},
+		{
+			in: PropertyPathFilter{[]string{"a"}, "==", math.NaN()},
+			want: &pb.StructuredQuery_Filter{FilterType: &pb.StructuredQuery_Filter_UnaryFilter{
+				UnaryFilter: &pb.StructuredQuery_UnaryFilter{
+					OperandType: &pb.StructuredQuery_UnaryFilter_Field{
+						Field: &pb.StructuredQuery_FieldReference{FieldPath: "a"},
+					},
+					Op: pb.StructuredQuery_UnaryFilter_IS_NAN,
+				},
+			}},
+		},
+		{
+			in: OrFilter{
+				Filters: []EntityFilter{
+					PropertyPathFilter{[]string{"a"}, ">", 5},
+					PropertyFilter{"a", "<=", 2},
+				},
+			},
+			want: &pb.StructuredQuery_Filter{
+				FilterType: &pb.StructuredQuery_Filter_CompositeFilter{
+					CompositeFilter: &pb.StructuredQuery_CompositeFilter{
+						Op: pb.StructuredQuery_CompositeFilter_OR,
+						Filters: []*pb.StructuredQuery_Filter{
+							{
+								FilterType: &pb.StructuredQuery_Filter_FieldFilter{
+									FieldFilter: &pb.StructuredQuery_FieldFilter{
+										Field: &pb.StructuredQuery_FieldReference{FieldPath: "a"},
+										Op:    pb.StructuredQuery_FieldFilter_GREATER_THAN,
+										Value: intval(5),
+									},
+								},
+							},
+							{
+								FilterType: &pb.StructuredQuery_Filter_FieldFilter{
+									FieldFilter: &pb.StructuredQuery_FieldFilter{
+										Field: &pb.StructuredQuery_FieldReference{FieldPath: "a"},
+										Op:    pb.StructuredQuery_FieldFilter_LESS_THAN_OR_EQUAL,
+										Value: intval(2),
+									},
+								},
+							},
+						},
+					},
+				},
+			},
+		},
+		{
+			in: AndFilter{
+				Filters: []EntityFilter{
+					PropertyPathFilter{[]string{"a"}, ">", 5},
+					PropertyFilter{"a", "<=", 10},
+				},
+			},
+			want: &pb.StructuredQuery_Filter{
+				FilterType: &pb.StructuredQuery_Filter_CompositeFilter{
+					CompositeFilter: &pb.StructuredQuery_CompositeFilter{
+						Op: pb.StructuredQuery_CompositeFilter_AND,
+						Filters: []*pb.StructuredQuery_Filter{
+							{
+								FilterType: &pb.StructuredQuery_Filter_FieldFilter{
+									FieldFilter: &pb.StructuredQuery_FieldFilter{
+										Field: &pb.StructuredQuery_FieldReference{FieldPath: "a"},
+										Op:    pb.StructuredQuery_FieldFilter_GREATER_THAN,
+										Value: intval(5),
+									},
+								},
+							},
+							{
+								FilterType: &pb.StructuredQuery_Filter_FieldFilter{
+									FieldFilter: &pb.StructuredQuery_FieldFilter{
+										Field: &pb.StructuredQuery_FieldReference{FieldPath: "a"},
+										Op:    pb.StructuredQuery_FieldFilter_LESS_THAN_OR_EQUAL,
+										Value: intval(10),
+									},
+								},
+							},
+						},
+					},
+				},
+			},
+		},
+		{
+			in: OrFilter{
+				Filters: []EntityFilter{
+					PropertyPathFilter{[]string{"b"}, "==", 15},
+					AndFilter{
+						Filters: []EntityFilter{
+							PropertyPathFilter{[]string{"a"}, ">", 5},
+							PropertyFilter{"a", "<=", 12},
+						},
+					},
+				},
+			},
+			want: &pb.StructuredQuery_Filter{
+				FilterType: &pb.StructuredQuery_Filter_CompositeFilter{
+					CompositeFilter: &pb.StructuredQuery_CompositeFilter{
+						Op: pb.StructuredQuery_CompositeFilter_OR,
+						Filters: []*pb.StructuredQuery_Filter{
+							{
+								FilterType: &pb.StructuredQuery_Filter_FieldFilter{
+									FieldFilter: &pb.StructuredQuery_FieldFilter{
+										Field: &pb.StructuredQuery_FieldReference{FieldPath: "b"},
+										Op:    pb.StructuredQuery_FieldFilter_EQUAL,
+										Value: intval(15),
+									},
+								},
+							},
+							{
+								FilterType: &pb.StructuredQuery_Filter_CompositeFilter{
+									CompositeFilter: &pb.StructuredQuery_CompositeFilter{
+										Op: pb.StructuredQuery_CompositeFilter_AND,
+										Filters: []*pb.StructuredQuery_Filter{
+											{
+												FilterType: &pb.StructuredQuery_Filter_FieldFilter{
+													FieldFilter: &pb.StructuredQuery_FieldFilter{
+														Field: &pb.StructuredQuery_FieldReference{FieldPath: "a"},
+														Op:    pb.StructuredQuery_FieldFilter_GREATER_THAN,
+														Value: intval(5),
+													},
+												},
+											},
+											{
+												FilterType: &pb.StructuredQuery_Filter_FieldFilter{
+													FieldFilter: &pb.StructuredQuery_FieldFilter{
+														Field: &pb.StructuredQuery_FieldReference{FieldPath: "a"},
+														Op:    pb.StructuredQuery_FieldFilter_LESS_THAN_OR_EQUAL,
+														Value: intval(12),
+													},
+												},
+											},
+										},
+									},
+								},
+							},
+						},
+					},
+				},
+			},
+		},
 	} {
 		got, err := test.in.toProto()
 		if err != nil {
@@ -74,9 +237,16 @@ func TestFilterToProto(t *testing.T) {
 	}
 }
 
-func TestQueryToProto(t *testing.T) {
+type toProtoScenario struct {
+	desc string
+	in   Query
+	want *pb.StructuredQuery
+}
+
+// Creates protos used to test toProto, FromProto, ToProto funcs.
+func createTestScenarios(t *testing.T) []toProtoScenario {
 	filtr := func(path []string, op string, val interface{}) *pb.StructuredQuery_Filter {
-		f, err := filter{path, op, val}.toProto()
+		f, err := PropertyPathFilter{path, op, val}.toProto()
 		if err != nil {
 			t.Fatal(err)
 		}
@@ -92,11 +262,8 @@ func TestQueryToProto(t *testing.T) {
 			Fields: map[string]*pb.Value{"a": intval(7), "b": intval(8), "c": arrayval(intval(1), intval(2))},
 		},
 	}
-	for _, test := range []struct {
-		desc string
-		in   Query
-		want *pb.StructuredQuery
-	}{
+
+	return []toProtoScenario{
 		{
 			desc: "q.Select()",
 			in:   q.Select(),
@@ -143,11 +310,21 @@ func TestQueryToProto(t *testing.T) {
 			in:   q.Where("a", "==", float32(math.NaN())),
 			want: &pb.StructuredQuery{Where: filtr([]string{"a"}, "==", math.NaN())},
 		},
+		{
+			desc: `q.Where("a", "!=", 3)`,
+			in:   q.Where("a", "!=", 3),
+			want: &pb.StructuredQuery{Where: filtr([]string{"a"}, "!=", 3)},
+		},
 		{
 			desc: `q.Where("a", "in", []int{7, 8})`,
 			in:   q.Where("a", "in", []int{7, 8}),
 			want: &pb.StructuredQuery{Where: filtr([]string{"a"}, "in", []int{7, 8})},
 		},
+		{
+			desc: `q.Where("a", "not-in", []int{9})`,
+			in:   q.Where("a", "not-in", []int{9}),
+			want: &pb.StructuredQuery{Where: filtr([]string{"a"}, "not-in", []int{9})},
+		},
 		{
 			desc: `q.Where("c", "array-contains", 1)`,
 			in:   q.Where("c", "array-contains", 1),
@@ -164,7 +341,7 @@ func TestQueryToProto(t *testing.T) {
 			want: &pb.StructuredQuery{
 				Where: &pb.StructuredQuery_Filter{
 					FilterType: &pb.StructuredQuery_Filter_CompositeFilter{
-						&pb.StructuredQuery_CompositeFilter{
+						CompositeFilter: &pb.StructuredQuery_CompositeFilter{
 							Op: pb.StructuredQuery_CompositeFilter_AND,
 							Filters: []*pb.StructuredQuery_Filter{
 								filtr([]string{"a"}, ">", 5), filtr([]string{"b"}, "<", "foo"),
@@ -174,6 +351,54 @@ func TestQueryToProto(t *testing.T) {
 				},
 			},
 		},
+		{
+			desc: `q.WhereEntity(AndFilter({"a", ">", 5}, {"b", "<", "foo"}))`,
+			in: q.WhereEntity(
+				AndFilter{
+					Filters: []EntityFilter{
+						PropertyFilter{
+							Path:     "a",
+							Operator: ">",
+							Value:    5,
+						},
+						PropertyFilter{
+							Path:     "b",
+							Operator: "<",
+							Value:    "foo",
+						},
+					},
+				},
+			),
+			want: &pb.StructuredQuery{
+				Where: &pb.StructuredQuery_Filter{
+					FilterType: &pb.StructuredQuery_Filter_CompositeFilter{
+						CompositeFilter: &pb.StructuredQuery_CompositeFilter{
+							Op: pb.StructuredQuery_CompositeFilter_AND,
+							Filters: []*pb.StructuredQuery_Filter{
+								{
+									FilterType: &pb.StructuredQuery_Filter_FieldFilter{
+										FieldFilter: &pb.StructuredQuery_FieldFilter{
+											Field: &pb.StructuredQuery_FieldReference{FieldPath: "a"},
+											Op:    pb.StructuredQuery_FieldFilter_GREATER_THAN,
+											Value: intval(5),
+										},
+									},
+								},
+								{
+									FilterType: &pb.StructuredQuery_Filter_FieldFilter{
+										FieldFilter: &pb.StructuredQuery_FieldFilter{
+											Field: &pb.StructuredQuery_FieldReference{FieldPath: "b"},
+											Op:    pb.StructuredQuery_FieldFilter_LESS_THAN,
+											Value: strval("foo"),
+										},
+									},
+								},
+							},
+						},
+					},
+				},
+			},
+		},
 		{
 			desc: `  q.WherePath([]string{"/", "*"}, ">", 5)`,
 			in:   q.WherePath([]string{"/", "*"}, ">", 5),
@@ -195,7 +420,7 @@ func TestQueryToProto(t *testing.T) {
 			in:   q.Offset(2).Limit(3),
 			want: &pb.StructuredQuery{
 				Offset: 2,
-				Limit:  &wrappers.Int32Value{Value: 3},
+				Limit:  &wrapperspb.Int32Value{Value: 3},
 			},
 		},
 		{
@@ -203,7 +428,7 @@ func TestQueryToProto(t *testing.T) {
 			in:   q.Offset(2).Limit(3).Limit(4).Offset(5), // last wins
 			want: &pb.StructuredQuery{
 				Offset: 5,
-				Limit:  &wrappers.Int32Value{Value: 4},
+				Limit:  &wrapperspb.Int32Value{Value: 4},
 			},
 		},
 		{
@@ -406,7 +631,7 @@ func TestQueryToProto(t *testing.T) {
 			want: &pb.StructuredQuery{
 				Where: &pb.StructuredQuery_Filter{
 					FilterType: &pb.StructuredQuery_Filter_CompositeFilter{
-						&pb.StructuredQuery_CompositeFilter{
+						CompositeFilter: &pb.StructuredQuery_CompositeFilter{
 							Op: pb.StructuredQuery_CompositeFilter_AND,
 							Filters: []*pb.StructuredQuery_Filter{
 								filtr([]string{"b"}, "==", 1),
@@ -425,11 +650,14 @@ func TestQueryToProto(t *testing.T) {
 				},
 			},
 		},
-	} {
+	}
+}
+
+func TestQueryToProto(t *testing.T) {
+	for _, test := range createTestScenarios(t) {
 		got, err := test.in.toProto()
 		if err != nil {
 			t.Fatalf("%s: %v", test.desc, err)
-			continue
 		}
 		test.want.From = []*pb.StructuredQuery_CollectionSelector{{CollectionId: "C"}}
 		if !testEqual(got, test.want) {
@@ -438,8 +666,39 @@ func TestQueryToProto(t *testing.T) {
 	}
 }
 
+// Convert a Query to a Proto and back again verifying roundtripping
+func TestQueryFromProtoRoundTrip(t *testing.T) {
+	c := &Client{projectID: "P", databaseID: "DB"}
+
+	for _, test := range createTestScenarios(t) {
+		fmt.Println(test.desc)
+		proto, err := test.in.Serialize()
+		if err != nil {
+			t.Fatalf("%s: %v", test.desc, err)
+		}
+		fmt.Printf("proto: %v\n", proto)
+		got, err := Query{c: c}.Deserialize(proto)
+		if err != nil {
+			t.Fatalf("%s: %v", test.desc, err)
+		}
+
+		want := test.in
+		gotProto, err := got.Serialize()
+		fmt.Println(gotProto)
+		if err != nil {
+			t.Fatalf("%s: %v", test.desc, err)
+		}
+
+		// Compare protos before and after taking to a query. proto -> query -> proto.
+		if diff := cmp.Diff(gotProto, proto, protocmp.Transform()); diff != "" {
+			t.Errorf("%s:\ngot\n%v\nwant\n%v\ndiff\n%v", test.desc, pretty.Value(got), pretty.Value(want), diff)
+		}
+	}
+}
+
 func fref1(s string) *pb.StructuredQuery_FieldReference {
-	return fref([]string{s})
+	ref, _ := fref([]string{s})
+	return ref
 }
 
 func TestQueryToProtoErrors(t *testing.T) {
@@ -456,10 +715,26 @@ func TestQueryToProtoErrors(t *testing.T) {
 	q := coll.Query
 	for i, query := range []Query{
 		{},                                     // no collection ID
-		q.Where("x", "!=", 1),                  // invalid operator
+		q.Where("x", "<>", 1),                  // invalid operator
 		q.Where("~", ">", 1),                   // invalid path
 		q.WherePath([]string{"*", ""}, ">", 1), // invalid path
-		q.StartAt(1),                           // no OrderBy
+		q.WhereEntity( // invalid nested filters
+			AndFilter{
+				Filters: []EntityFilter{
+					PropertyFilter{
+						Path:     "x",
+						Operator: "<>",
+						Value:    1,
+					},
+					PropertyFilter{
+						Path:     "~",
+						Operator: ">",
+						Value:    1,
+					},
+				},
+			},
+		),
+		q.StartAt(1), // no OrderBy
 		q.StartAt(2).OrderBy("x", Asc).OrderBy("y", Desc), // wrong # OrderBy
 		q.Select("*"),                         // invalid path
 		q.SelectPaths([]string{"/", "", "~"}), // invalid path
@@ -558,13 +833,16 @@ func TestQueryFromCollectionRef(t *testing.T) {
 	c := &Client{projectID: "P", databaseID: "D"}
 	coll := c.Collection("C")
 	got := coll.Select("x").Offset(8)
+	ref, _ := fref(FieldPath{"x"})
 	want := Query{
 		c:            c,
 		parentPath:   c.path() + "/documents",
 		path:         "projects/P/databases/D/documents/C",
 		collectionID: "C",
-		selection:    []FieldPath{{"x"}},
-		offset:       8,
+		selection: []*pb.StructuredQuery_FieldReference{
+			ref,
+		},
+		offset: 8,
 	}
 	if !testEqual(got, want) {
 		t.Fatalf("\ngot  %+v, \nwant %+v", got, want)
@@ -869,3 +1147,215 @@ func (b byQuery) Less(i, j int) bool {
 	}
 	return c < 0
 }
+
+func TestAggregationQuery(t *testing.T) {
+	ctx := context.Background()
+	c, srv, cleanup := newMock(t)
+	defer cleanup()
+
+	srv.addRPC(nil, []interface{}{
+		&pb.RunAggregationQueryResponse{
+			Result: &pb.AggregationResult{
+				AggregateFields: map[string]*pb.Value{
+					"testAlias": intval(1),
+				},
+			},
+		},
+	})
+
+	q := c.Collection("coll1").Where("f", "==", 2)
+	ar, err := q.NewAggregationQuery().WithCount("testAlias").Get(ctx)
+	if err != nil {
+		t.Fatal(err)
+	}
+
+	count, ok := ar["testAlias"]
+	if !ok {
+		t.Errorf("aggregation query key not found")
+	}
+
+	cv := count.(*pb.Value)
+	if cv.GetIntegerValue() != 1 {
+		t.Errorf("got: %v\nwant: %v\n; result: %v\n", cv.GetIntegerValue(), 1, count)
+	}
+}
+
+func TestWithSum(t *testing.T) {
+	ctx := context.Background()
+	sumAlias := "sum"
+	c, srv, cleanup := newMock(t)
+	defer cleanup()
+
+	srv.addRPC(nil, []interface{}{
+		&pb.RunAggregationQueryResponse{
+			Result: &pb.AggregationResult{
+				AggregateFields: map[string]*pb.Value{
+					"sum": intval(1),
+				},
+			},
+		},
+	})
+
+	testcases := []struct {
+		desc    string
+		path    string
+		wantErr bool
+	}{
+		{
+			desc:    "Invalid path",
+			path:    "path*",
+			wantErr: true,
+		},
+		{
+			desc:    "Valid path",
+			path:    "path",
+			wantErr: false,
+		},
+	}
+	for _, tc := range testcases {
+
+		query := c.Collection("C")
+		aggQuery := query.NewAggregationQuery().WithSum(tc.path, sumAlias)
+		_, err := aggQuery.Get(ctx)
+		if err == nil && tc.wantErr {
+			t.Fatalf("%s: got nil wanted error", tc.desc)
+		} else if err != nil && !tc.wantErr {
+			t.Fatalf("%s: got %v, want nil", tc.desc, err)
+		}
+	}
+}
+
+func TestWithSumPath(t *testing.T) {
+	ctx := context.Background()
+	sumAlias := "sum"
+	c, srv, cleanup := newMock(t)
+	defer cleanup()
+
+	srv.addRPC(nil, []interface{}{
+		&pb.RunAggregationQueryResponse{
+			Result: &pb.AggregationResult{
+				AggregateFields: map[string]*pb.Value{
+					"sum": intval(1),
+				},
+			},
+		},
+	})
+
+	testcases := []struct {
+		desc      string
+		fieldPath FieldPath
+		wantErr   bool
+	}{
+		{
+			desc:      "Invalid path",
+			fieldPath: []string{},
+			wantErr:   true,
+		},
+		{
+			desc:      "Valid path",
+			fieldPath: []string{"path"},
+			wantErr:   false,
+		},
+	}
+	for _, tc := range testcases {
+
+		query := c.Collection("C")
+		aggQuery := query.NewAggregationQuery().WithSumPath(tc.fieldPath, sumAlias)
+		_, err := aggQuery.Get(ctx)
+		if err == nil && tc.wantErr {
+			t.Fatalf("%s: got nil wanted error", tc.desc)
+		} else if err != nil && !tc.wantErr {
+			t.Fatalf("%s: got %v, want nil", tc.desc, err)
+		}
+	}
+}
+
+func TestWithAvg(t *testing.T) {
+	ctx := context.Background()
+	avgAlias := "avg"
+	c, srv, cleanup := newMock(t)
+	defer cleanup()
+
+	srv.addRPC(nil, []interface{}{
+		&pb.RunAggregationQueryResponse{
+			Result: &pb.AggregationResult{
+				AggregateFields: map[string]*pb.Value{
+					"avg": intval(1),
+				},
+			},
+		},
+	})
+
+	testcases := []struct {
+		desc    string
+		path    string
+		wantErr bool
+	}{
+		{
+			desc:    "Invalid path",
+			path:    "path*",
+			wantErr: true,
+		},
+		{
+			desc:    "Valid path",
+			path:    "path",
+			wantErr: false,
+		},
+	}
+	for _, tc := range testcases {
+
+		query := c.Collection("C")
+		aggQuery := query.NewAggregationQuery().WithAvg(tc.path, avgAlias)
+		_, err := aggQuery.Get(ctx)
+		if err == nil && tc.wantErr {
+			t.Fatalf("%s: got nil wanted error", tc.desc)
+		} else if err != nil && !tc.wantErr {
+			t.Fatalf("%s: got %v, want nil", tc.desc, err)
+		}
+	}
+}
+
+func TestWithAvgPath(t *testing.T) {
+	ctx := context.Background()
+	avgAlias := "avg"
+	c, srv, cleanup := newMock(t)
+	defer cleanup()
+
+	srv.addRPC(nil, []interface{}{
+		&pb.RunAggregationQueryResponse{
+			Result: &pb.AggregationResult{
+				AggregateFields: map[string]*pb.Value{
+					"avg": intval(1),
+				},
+			},
+		},
+	})
+
+	testcases := []struct {
+		desc      string
+		fieldPath FieldPath
+		wantErr   bool
+	}{
+		{
+			desc:      "Invalid path",
+			fieldPath: []string{},
+			wantErr:   true,
+		},
+		{
+			desc:      "Valid path",
+			fieldPath: []string{"path"},
+			wantErr:   false,
+		},
+	}
+	for _, tc := range testcases {
+
+		query := c.Collection("C")
+		aggQuery := query.NewAggregationQuery().WithAvgPath(tc.fieldPath, avgAlias)
+		_, err := aggQuery.Get(ctx)
+		if err == nil && tc.wantErr {
+			t.Fatalf("%s: got nil wanted error", tc.desc)
+		} else if err != nil && !tc.wantErr {
+			t.Fatalf("%s: got %v, want nil", tc.desc, err)
+		}
+	}
+}
