@@ -1,5 +1,5 @@
 /*
-Copyright 2017 Google Inc. All Rights Reserved.
+Copyright 2017 Google LLC
 
 Licensed under the Apache License, Version 2.0 (the "License");
 you may not use this file except in compliance with the License.
@@ -19,11 +19,10 @@ package testing
 import (
 	"fmt"
 
-	"github.com/golang/protobuf/proto"
+	"google.golang.org/protobuf/proto"
 )
 
-// Compare two payloads, assuming they are both proto.Messages
-// or both strings.
+// PayloadEqual compares two payloads, assuming they are both proto.Messages or both strings.
 func PayloadEqual(a, b interface{}) bool {
 	if a == nil && b == nil {
 		return true
