@@ -30,29 +30,35 @@ import (
 	"encoding/json"
 	"errors"
 	"fmt"
+	"io"
 	"log"
 	"net/http"
-	"os"
 	"regexp"
+	"runtime"
 	"strconv"
 	"strings"
 	"sync"
 	"time"
 	"unicode/utf8"
 
-	"cloud.google.com/go/compute/metadata"
-	"cloud.google.com/go/internal/version"
+	"go.opentelemetry.io/otel/trace"
+
 	vkit "cloud.google.com/go/logging/apiv2"
+	logpb "cloud.google.com/go/logging/apiv2/loggingpb"
 	"cloud.google.com/go/logging/internal"
-	"github.com/golang/protobuf/proto"
-	"github.com/golang/protobuf/ptypes"
-	structpb "github.com/golang/protobuf/ptypes/struct"
-	tspb "github.com/golang/protobuf/ptypes/timestamp"
+	gax "github.com/googleapis/gax-go/v2"
 	"google.golang.org/api/option"
 	"google.golang.org/api/support/bundler"
 	mrpb "google.golang.org/genproto/googleapis/api/monitoredres"
 	logtypepb "google.golang.org/genproto/googleapis/logging/type"
-	logpb "google.golang.org/genproto/googleapis/logging/v2"
+	"google.golang.org/grpc/codes"
+	"google.golang.org/grpc/status"
+	"google.golang.org/protobuf/encoding/protojson"
+	"google.golang.org/protobuf/proto"
+	"google.golang.org/protobuf/types/known/anypb"
+	"google.golang.org/protobuf/types/known/durationpb"
+	structpb "google.golang.org/protobuf/types/known/structpb"
+	"google.golang.org/protobuf/types/known/timestamppb"
 )
 
 const (
@@ -78,7 +84,10 @@ const (
 	DefaultEntryCountThreshold = 1000
 
 	// DefaultEntryByteThreshold is the default value for the EntryByteThreshold LoggerOption.
-	DefaultEntryByteThreshold = 1 << 20 // 1MiB
+	DefaultEntryByteThreshold = 1 << 23 // 8MiB
+
+	// DefaultBundleByteLimit is the default value for the BundleByteLimit LoggerOption.
+	DefaultBundleByteLimit = 9437184 // 9.5 MiB
 
 	// DefaultBufferedByteLimit is the default value for the BufferedByteLimit LoggerOption.
 	DefaultBufferedByteLimit = 1 << 30 // 1GiB
@@ -88,18 +97,38 @@ const (
 	// timeout is to allow clients to degrade gracefully if underlying logging
 	// service is temporarily impaired for some reason.
 	defaultWriteTimeout = 10 * time.Minute
+
+	// Part of the error message when the payload contains invalid UTF-8 characters.
+	utfErrorString = "string field contains invalid UTF-8"
+
+	// DetectProjectID is a sentinel value that instructs NewClient to detect the
+	// project ID. It is given in place of the projectID argument. NewClient will
+	// use the project ID from the given credentials or the default credentials
+	// (https://developers.google.com/accounts/docs/application-default-credentials)
+	// if no credentials were provided. When providing credentials, not all
+	// options will allow NewClient to extract the project ID. Specifically a JWT
+	// does not have the project ID encoded.
+	DetectProjectID = "*detect-project-id*"
 )
 
-// For testing:
-var now = time.Now
+var (
+	// ErrRedirectProtoPayloadNotSupported is returned when Logger is configured to redirect output and
+	// tries to redirect logs with protobuf payload.
+	ErrRedirectProtoPayloadNotSupported = errors.New("printEntryToStdout: cannot find valid payload")
 
-// ErrOverflow signals that the number of buffered entries for a Logger
-// exceeds its BufferLimit.
-var ErrOverflow = bundler.ErrOverflow
+	// For testing:
+	now                    = time.Now
+	toLogEntryInternal     = toLogEntryInternalImpl
+	detectResourceInternal = detectResource
 
-// ErrOversizedEntry signals that an entry's size exceeds the maximum number of
-// bytes that will be sent in a single call to the logging service.
-var ErrOversizedEntry = bundler.ErrOversizedItem
+	// ErrOverflow signals that the number of buffered entries for a Logger
+	// exceeds its BufferLimit.
+	ErrOverflow = bundler.ErrOverflow
+
+	// ErrOversizedEntry signals that an entry's size exceeds the maximum number of
+	// bytes that will be sent in a single call to the logging service.
+	ErrOversizedEntry = bundler.ErrOversizedItem
+)
 
 // Client is a Logging client. A Client is associated with a single Cloud project.
 type Client struct {
@@ -128,28 +157,33 @@ type Client struct {
 
 // NewClient returns a new logging client associated with the provided parent.
 // A parent can take any of the following forms:
-//    projects/PROJECT_ID
-//    folders/FOLDER_ID
-//    billingAccounts/ACCOUNT_ID
-//    organizations/ORG_ID
-// for backwards compatibility, a string with no '/' is also allowed and is interpreted
+//
+//	projects/PROJECT_ID
+//	folders/FOLDER_ID
+//	billingAccounts/ACCOUNT_ID
+//	organizations/ORG_ID
+//
+// For backwards compatibility, a string with no '/' is also allowed and is interpreted
 // as a project ID.
 //
+// If logging.DetectProjectId is provided as the parent, the parent will be interpreted as a project
+// ID, and its value will be inferred from the environment.
+//
 // By default NewClient uses WriteScope. To use a different scope, call
 // NewClient using a WithScopes option (see https://godoc.org/google.golang.org/api/option#WithScopes).
 func NewClient(ctx context.Context, parent string, opts ...option.ClientOption) (*Client, error) {
-	if !strings.ContainsRune(parent, '/') {
-		parent = "projects/" + parent
+	parent, err := makeParent(parent)
+	if err != nil {
+		return nil, err
 	}
 	opts = append([]option.ClientOption{
-		option.WithEndpoint(internal.ProdAddr),
 		option.WithScopes(WriteScope),
 	}, opts...)
 	c, err := vkit.NewClient(ctx, opts...)
 	if err != nil {
 		return nil, err
 	}
-	c.SetGoogleClientInfo("gccl", version.Repo)
+	c.SetGoogleClientInfo("gccl", internal.Version)
 	client := &Client{
 		client:  c,
 		parent:  parent,
@@ -174,20 +208,29 @@ func NewClient(ctx context.Context, parent string, opts ...option.ClientOption)
 	return client, nil
 }
 
-var unixZeroTimestamp *tspb.Timestamp
-
-func init() {
-	var err error
-	unixZeroTimestamp, err = ptypes.TimestampProto(time.Unix(0, 0))
-	if err != nil {
-		panic(err)
+func makeParent(parent string) (string, error) {
+	if !strings.ContainsRune(parent, '/') {
+		if parent == DetectProjectID {
+			resource := detectResourceInternal()
+			if resource == nil {
+				return parent, fmt.Errorf("could not determine project ID from environment")
+			}
+			parent = resource.Labels["project_id"]
+		}
+		return "projects/" + parent, nil
 	}
+	prefix := strings.Split(parent, "/")[0]
+	if prefix != "projects" && prefix != "folders" && prefix != "billingAccounts" && prefix != "organizations" {
+		return parent, fmt.Errorf("parent parameter must start with 'projects/' 'folders/' 'billingAccounts/' or 'organizations/'")
+	}
+	return parent, nil
 }
 
 // Ping reports whether the client's connection to the logging service and the
 // authentication configuration are valid. To accomplish this, Ping writes a
 // log entry "ping" to a log named "ping".
 func (c *Client) Ping(ctx context.Context) error {
+	unixZeroTimestamp := timestamppb.New(time.Unix(0, 0))
 	ent := &logpb.LogEntry{
 		Payload:   &logpb.LogEntry_TextPayload{TextPayload: "ping"},
 		Timestamp: unixZeroTimestamp, // Identical timestamps and insert IDs are both
@@ -218,7 +261,7 @@ func (c *Client) extractErrorInfo() error {
 	var err error
 	c.mu.Lock()
 	if c.lastErr != nil {
-		err = fmt.Errorf("saw %d errors; last: %v", c.nErrs, c.lastErr)
+		err = fmt.Errorf("saw %d errors; last: %w", c.nErrs, c.lastErr)
 		c.nErrs = 0
 		c.lastErr = nil
 	}
@@ -235,230 +278,63 @@ type Logger struct {
 	bundler    *bundler.Bundler
 
 	// Options
-	commonResource *mrpb.MonitoredResource
-	commonLabels   map[string]string
-	ctxFunc        func() (context.Context, func())
+	commonResource         *mrpb.MonitoredResource
+	commonLabels           map[string]string
+	ctxFunc                func() (context.Context, func())
+	populateSourceLocation int
+	partialSuccess         bool
+	redirectOutputWriter   io.Writer
 }
 
-// A LoggerOption is a configuration option for a Logger.
-type LoggerOption interface {
-	set(*Logger)
+type loggerRetryer struct {
+	defaultRetryer gax.Retryer
 }
 
-// CommonResource sets the monitored resource associated with all log entries
-// written from a Logger. If not provided, the resource is automatically
-// detected based on the running environment (on GCE and GAE Standard only).
-// This value can be overridden per-entry by setting an Entry's Resource field.
-func CommonResource(r *mrpb.MonitoredResource) LoggerOption { return commonResource{r} }
-
-type commonResource struct{ *mrpb.MonitoredResource }
-
-func (r commonResource) set(l *Logger) { l.commonResource = r.MonitoredResource }
-
-var detectedResource struct {
-	pb   *mrpb.MonitoredResource
-	once sync.Once
-}
-
-func detectGCEResource() *mrpb.MonitoredResource {
-	projectID, err := metadata.ProjectID()
-	if err != nil {
-		return nil
-	}
-	id, err := metadata.InstanceID()
-	if err != nil {
-		return nil
-	}
-	zone, err := metadata.Zone()
-	if err != nil {
-		return nil
-	}
-	name, err := metadata.InstanceName()
-	if err != nil {
-		return nil
-	}
-	return &mrpb.MonitoredResource{
-		Type: "gce_instance",
-		Labels: map[string]string{
-			"project_id":    projectID,
-			"instance_id":   id,
-			"instance_name": name,
-			"zone":          zone,
-		},
-	}
-}
-
-func detectGAEResource() *mrpb.MonitoredResource {
-	return &mrpb.MonitoredResource{
-		Type: "gae_app",
-		Labels: map[string]string{
-			"project_id":  os.Getenv("GOOGLE_CLOUD_PROJECT"),
-			"module_id":   os.Getenv("GAE_SERVICE"),
-			"version_id":  os.Getenv("GAE_VERSION"),
-			"instance_id": os.Getenv("GAE_INSTANCE"),
-			"runtime":     os.Getenv("GAE_RUNTIME"),
-		},
-	}
-}
-
-func detectResource() *mrpb.MonitoredResource {
-	detectedResource.once.Do(func() {
-		switch {
-		// GAE needs to come first, as metadata.OnGCE() is actually true on GAE
-		// Second Gen runtimes.
-		case os.Getenv("GAE_ENV") == "standard":
-			detectedResource.pb = detectGAEResource()
-		case metadata.OnGCE():
-			detectedResource.pb = detectGCEResource()
-		}
+func newLoggerRetryer() gax.Retryer {
+	// Copied from CallOptions.WriteLogEntries in apiv2/logging_client.go.
+	d := gax.OnCodes([]codes.Code{
+		codes.DeadlineExceeded,
+		codes.Internal,
+		codes.Unavailable,
+	}, gax.Backoff{
+		Initial:    100 * time.Millisecond,
+		Max:        60000 * time.Millisecond,
+		Multiplier: 1.30,
 	})
-	return detectedResource.pb
+
+	r := &loggerRetryer{defaultRetryer: d}
+	return r
 }
 
-var resourceInfo = map[string]struct{ rtype, label string }{
-	"organizations":   {"organization", "organization_id"},
-	"folders":         {"folder", "folder_id"},
-	"projects":        {"project", "project_id"},
-	"billingAccounts": {"billing_account", "account_id"},
-}
-
-func monitoredResource(parent string) *mrpb.MonitoredResource {
-	parts := strings.SplitN(parent, "/", 2)
-	if len(parts) != 2 {
-		return globalResource(parent)
-	}
-	info, ok := resourceInfo[parts[0]]
+func (r *loggerRetryer) Retry(err error) (pause time.Duration, shouldRetry bool) {
+	s, ok := status.FromError(err)
 	if !ok {
-		return globalResource(parts[1])
+		return r.defaultRetryer.Retry(err)
 	}
-	return &mrpb.MonitoredResource{
-		Type:   info.rtype,
-		Labels: map[string]string{info.label: parts[1]},
+	if strings.Contains(s.Message(), utfErrorString) {
+		return 0, false
 	}
+	return r.defaultRetryer.Retry(err)
 }
 
-func globalResource(projectID string) *mrpb.MonitoredResource {
-	return &mrpb.MonitoredResource{
-		Type: "global",
-		Labels: map[string]string{
-			"project_id": projectID,
-		},
-	}
-}
-
-// CommonLabels are labels that apply to all log entries written from a Logger,
-// so that you don't have to repeat them in each log entry's Labels field. If
-// any of the log entries contains a (key, value) with the same key that is in
-// CommonLabels, then the entry's (key, value) overrides the one in
-// CommonLabels.
-func CommonLabels(m map[string]string) LoggerOption { return commonLabels(m) }
-
-type commonLabels map[string]string
-
-func (c commonLabels) set(l *Logger) { l.commonLabels = c }
-
-// ConcurrentWriteLimit determines how many goroutines will send log entries to the
-// underlying service. The default is 1. Set ConcurrentWriteLimit to a higher value to
-// increase throughput.
-func ConcurrentWriteLimit(n int) LoggerOption { return concurrentWriteLimit(n) }
-
-type concurrentWriteLimit int
-
-func (c concurrentWriteLimit) set(l *Logger) { l.bundler.HandlerLimit = int(c) }
-
-// DelayThreshold is the maximum amount of time that an entry should remain
-// buffered in memory before a call to the logging service is triggered. Larger
-// values of DelayThreshold will generally result in fewer calls to the logging
-// service, while increasing the risk that log entries will be lost if the
-// process crashes.
-// The default is DefaultDelayThreshold.
-func DelayThreshold(d time.Duration) LoggerOption { return delayThreshold(d) }
-
-type delayThreshold time.Duration
-
-func (d delayThreshold) set(l *Logger) { l.bundler.DelayThreshold = time.Duration(d) }
-
-// EntryCountThreshold is the maximum number of entries that will be buffered
-// in memory before a call to the logging service is triggered. Larger values
-// will generally result in fewer calls to the logging service, while
-// increasing both memory consumption and the risk that log entries will be
-// lost if the process crashes.
-// The default is DefaultEntryCountThreshold.
-func EntryCountThreshold(n int) LoggerOption { return entryCountThreshold(n) }
-
-type entryCountThreshold int
-
-func (e entryCountThreshold) set(l *Logger) { l.bundler.BundleCountThreshold = int(e) }
-
-// EntryByteThreshold is the maximum number of bytes of entries that will be
-// buffered in memory before a call to the logging service is triggered. See
-// EntryCountThreshold for a discussion of the tradeoffs involved in setting
-// this option.
-// The default is DefaultEntryByteThreshold.
-func EntryByteThreshold(n int) LoggerOption { return entryByteThreshold(n) }
-
-type entryByteThreshold int
-
-func (e entryByteThreshold) set(l *Logger) { l.bundler.BundleByteThreshold = int(e) }
-
-// EntryByteLimit is the maximum number of bytes of entries that will be sent
-// in a single call to the logging service. ErrOversizedEntry is returned if an
-// entry exceeds EntryByteLimit. This option limits the size of a single RPC
-// payload, to account for network or service issues with large RPCs. If
-// EntryByteLimit is smaller than EntryByteThreshold, the latter has no effect.
-// The default is zero, meaning there is no limit.
-func EntryByteLimit(n int) LoggerOption { return entryByteLimit(n) }
-
-type entryByteLimit int
-
-func (e entryByteLimit) set(l *Logger) { l.bundler.BundleByteLimit = int(e) }
-
-// BufferedByteLimit is the maximum number of bytes that the Logger will keep
-// in memory before returning ErrOverflow. This option limits the total memory
-// consumption of the Logger (but note that each Logger has its own, separate
-// limit). It is possible to reach BufferedByteLimit even if it is larger than
-// EntryByteThreshold or EntryByteLimit, because calls triggered by the latter
-// two options may be enqueued (and hence occupying memory) while new log
-// entries are being added.
-// The default is DefaultBufferedByteLimit.
-func BufferedByteLimit(n int) LoggerOption { return bufferedByteLimit(n) }
-
-type bufferedByteLimit int
-
-func (b bufferedByteLimit) set(l *Logger) { l.bundler.BufferedByteLimit = int(b) }
-
-// ContextFunc is a function that will be called to obtain a context.Context for the
-// WriteLogEntries RPC executed in the background for calls to Logger.Log. The
-// default is a function that always returns context.Background. The second return
-// value of the function is a function to call after the RPC completes.
-//
-// The function is not used for calls to Logger.LogSync, since the caller can pass
-// in the context directly.
-//
-// This option is EXPERIMENTAL. It may be changed or removed.
-func ContextFunc(f func() (ctx context.Context, afterCall func())) LoggerOption {
-	return contextFunc(f)
-}
-
-type contextFunc func() (ctx context.Context, afterCall func())
-
-func (c contextFunc) set(l *Logger) { l.ctxFunc = c }
-
 // Logger returns a Logger that will write entries with the given log ID, such as
 // "syslog". A log ID must be less than 512 characters long and can only
 // include the following characters: upper and lower case alphanumeric
 // characters: [A-Za-z0-9]; and punctuation characters: forward-slash,
 // underscore, hyphen, and period.
 func (c *Client) Logger(logID string, opts ...LoggerOption) *Logger {
-	r := detectResource()
+	r := detectResourceInternal()
 	if r == nil {
 		r = monitoredResource(c.parent)
 	}
 	l := &Logger{
-		client:         c,
-		logName:        internal.LogPath(c.parent, logID),
-		commonResource: r,
-		ctxFunc:        func() (context.Context, func()) { return context.Background(), nil },
+		client:                 c,
+		logName:                internal.LogPath(c.parent, logID),
+		commonResource:         r,
+		ctxFunc:                func() (context.Context, func()) { return context.Background(), nil },
+		populateSourceLocation: DoNotPopulateSourceLocation,
+		partialSuccess:         false,
+		redirectOutputWriter:   nil,
 	}
 	l.bundler = bundler.NewBundler(&logpb.LogEntry{}, func(entries interface{}) {
 		l.writeLogEntries(entries.([]*logpb.LogEntry))
@@ -466,13 +342,15 @@ func (c *Client) Logger(logID string, opts ...LoggerOption) *Logger {
 	l.bundler.DelayThreshold = DefaultDelayThreshold
 	l.bundler.BundleCountThreshold = DefaultEntryCountThreshold
 	l.bundler.BundleByteThreshold = DefaultEntryByteThreshold
+	l.bundler.BundleByteLimit = DefaultBundleByteLimit
 	l.bundler.BufferedByteLimit = DefaultBufferedByteLimit
 	for _, opt := range opts {
 		opt.set(l)
 	}
 	l.stdLoggers = map[Severity]*log.Logger{}
 	for s := range severityName {
-		l.stdLoggers[s] = log.New(severityWriter{l, s}, "", 0)
+		e := Entry{Severity: s}
+		l.stdLoggers[s] = log.New(templateEntryWriter{l, &e}, "", 0)
 	}
 
 	c.loggers.Add(1)
@@ -486,16 +364,20 @@ func (c *Client) Logger(logID string, opts ...LoggerOption) *Logger {
 	return l
 }
 
-type severityWriter struct {
-	l *Logger
-	s Severity
+type templateEntryWriter struct {
+	l        *Logger
+	template *Entry
 }
 
-func (w severityWriter) Write(p []byte) (n int, err error) {
-	w.l.Log(Entry{
-		Severity: w.s,
-		Payload:  string(p),
-	})
+func (w templateEntryWriter) Write(p []byte) (n int, err error) {
+	e := *w.template
+	e.Payload = string(p)
+	// The second argument to logInternal() is how many frames to skip
+	// from the call stack when determining the source location. In the
+	// current implementation of log.Logger (i.e. Go's logging library)
+	// the Write() method is called 2 calls deep so we need to skip 3
+	// frames to account for the call to logInternal() itself.
+	w.l.logInternal(e, 3)
 	return len(p), nil
 }
 
@@ -566,6 +448,21 @@ func (v Severity) String() string {
 	return strconv.Itoa(int(v))
 }
 
+// UnmarshalJSON turns a string representation of severity into the type
+// Severity.
+func (v *Severity) UnmarshalJSON(data []byte) error {
+	var s string
+	var i int
+	if strErr := json.Unmarshal(data, &s); strErr == nil {
+		*v = ParseSeverity(s)
+	} else if intErr := json.Unmarshal(data, &i); intErr == nil {
+		*v = Severity(i)
+	} else {
+		return fmt.Errorf("%v; %v", strErr, intErr)
+	}
+	return nil
+}
+
 // ParseSeverity returns the Severity whose name equals s, ignoring case. It
 // returns Default if no Severity matches.
 func ParseSeverity(s string) Severity {
@@ -676,14 +573,20 @@ type HTTPRequest struct {
 	// validated with the origin server before being served from cache. This
 	// field is only meaningful if CacheHit is true.
 	CacheValidatedWithOriginServer bool
+
+	// CacheFillBytes is the number of HTTP response bytes inserted into cache. Set only when a cache fill was attempted.
+	CacheFillBytes int64
+
+	// CacheLookup tells whether or not a cache lookup was attempted.
+	CacheLookup bool
 }
 
-func fromHTTPRequest(r *HTTPRequest) *logtypepb.HttpRequest {
+func fromHTTPRequest(r *HTTPRequest) (*logtypepb.HttpRequest, error) {
 	if r == nil {
-		return nil
+		return nil, nil
 	}
 	if r.Request == nil {
-		panic("HTTPRequest must have a non-nil Request")
+		return nil, errors.New("logging: HTTPRequest must have a non-nil Request")
 	}
 	u := *r.Request.URL
 	u.Fragment = ""
@@ -699,11 +602,14 @@ func fromHTTPRequest(r *HTTPRequest) *logtypepb.HttpRequest {
 		Referer:                        r.Request.Referer(),
 		CacheHit:                       r.CacheHit,
 		CacheValidatedWithOriginServer: r.CacheValidatedWithOriginServer,
+		Protocol:                       r.Request.Proto,
+		CacheFillBytes:                 r.CacheFillBytes,
+		CacheLookup:                    r.CacheLookup,
 	}
 	if r.Latency != 0 {
-		pb.Latency = ptypes.DurationProto(r.Latency)
+		pb.Latency = durationpb.New(r.Latency)
 	}
-	return pb
+	return pb, nil
 }
 
 // fixUTF8 is a helper that fixes an invalid UTF-8 string by replacing
@@ -745,13 +651,13 @@ func toProtoStruct(v interface{}) (*structpb.Struct, error) {
 	} else {
 		jb, err = json.Marshal(v)
 		if err != nil {
-			return nil, fmt.Errorf("logging: json.Marshal: %v", err)
+			return nil, fmt.Errorf("logging: json.Marshal: %w", err)
 		}
 	}
 	var m map[string]interface{}
 	err = json.Unmarshal(jb, &m)
 	if err != nil {
-		return nil, fmt.Errorf("logging: json.Unmarshal: %v", err)
+		return nil, fmt.Errorf("logging: json.Unmarshal: %w", err)
 	}
 	return jsonMapToProtoStruct(m), nil
 }
@@ -783,7 +689,7 @@ func jsonValueToStructValue(v interface{}) *structpb.Value {
 		}
 		return &structpb.Value{Kind: &structpb.Value_ListValue{ListValue: &structpb.ListValue{Values: vals}}}
 	default:
-		panic(fmt.Sprintf("bad type %T for JSON value", v))
+		return &structpb.Value{Kind: &structpb.Value_NullValue{}}
 	}
 }
 
@@ -791,28 +697,56 @@ func jsonValueToStructValue(v interface{}) *structpb.Value {
 // and will block, it is intended primarily for debugging or critical errors.
 // Prefer Log for most uses.
 func (l *Logger) LogSync(ctx context.Context, e Entry) error {
-	ent, err := l.toLogEntry(e)
+	ent, err := toLogEntryInternal(e, l, l.client.parent, 1)
 	if err != nil {
 		return err
 	}
+	entries, hasInstrumentation := l.instrumentLogs([]*logpb.LogEntry{ent})
+	if l.redirectOutputWriter != nil {
+		for _, ent = range entries {
+			err = serializeEntryToWriter(ent, l.redirectOutputWriter)
+			if err != nil {
+				break
+			}
+		}
+		return err
+	}
 	_, err = l.client.client.WriteLogEntries(ctx, &logpb.WriteLogEntriesRequest{
-		LogName:  l.logName,
-		Resource: l.commonResource,
-		Labels:   l.commonLabels,
-		Entries:  []*logpb.LogEntry{ent},
+		LogName:        l.logName,
+		Resource:       l.commonResource,
+		Labels:         l.commonLabels,
+		Entries:        entries,
+		PartialSuccess: l.partialSuccess || hasInstrumentation,
 	})
 	return err
 }
 
 // Log buffers the Entry for output to the logging service. It never blocks.
 func (l *Logger) Log(e Entry) {
-	ent, err := l.toLogEntry(e)
+	l.logInternal(e, 1)
+}
+
+func (l *Logger) logInternal(e Entry, skipLevels int) {
+	ent, err := toLogEntryInternal(e, l, l.client.parent, skipLevels+1)
 	if err != nil {
 		l.client.error(err)
 		return
 	}
-	if err := l.bundler.Add(ent, proto.Size(ent)); err != nil {
-		l.client.error(err)
+
+	entries, _ := l.instrumentLogs([]*logpb.LogEntry{ent})
+	if l.redirectOutputWriter != nil {
+		for _, ent = range entries {
+			err = serializeEntryToWriter(ent, l.redirectOutputWriter)
+			if err != nil {
+				l.client.error(err)
+			}
+		}
+		return
+	}
+	for _, ent = range entries {
+		if err := l.bundler.Add(ent, proto.Size(ent)); err != nil {
+			l.client.error(err)
+		}
 	}
 }
 
@@ -828,16 +762,22 @@ func (l *Logger) Flush() error {
 }
 
 func (l *Logger) writeLogEntries(entries []*logpb.LogEntry) {
+	partialSuccess := l.partialSuccess
+	if len(entries) > 1 {
+		partialSuccess = partialSuccess || hasInstrumentation(entries)
+	}
 	req := &logpb.WriteLogEntriesRequest{
-		LogName:  l.logName,
-		Resource: l.commonResource,
-		Labels:   l.commonLabels,
-		Entries:  entries,
+		LogName:        l.logName,
+		Resource:       l.commonResource,
+		Labels:         l.commonLabels,
+		Entries:        entries,
+		PartialSuccess: partialSuccess,
 	}
 	ctx, afterCall := l.ctxFunc()
 	ctx, cancel := context.WithTimeout(ctx, defaultWriteTimeout)
 	defer cancel()
-	_, err := l.client.client.WriteLogEntries(ctx, req)
+
+	_, err := l.client.client.WriteLogEntries(ctx, req, gax.WithRetry(newLoggerRetryer))
 	if err != nil {
 		l.client.error(err)
 	}
@@ -853,38 +793,141 @@ func (l *Logger) writeLogEntries(entries []*logpb.LogEntry) {
 // (for example by calling SetFlags or SetPrefix).
 func (l *Logger) StandardLogger(s Severity) *log.Logger { return l.stdLoggers[s] }
 
-var reCloudTraceContext = regexp.MustCompile(`([a-f\d]+)/([a-f\d]+);o=(\d)`)
+// StandardLoggerFromTemplate returns a Go Standard Logging API *log.Logger.
+//
+// The returned logger emits logs using logging.(*Logger).Log() with an entry
+// constructed from the provided template Entry struct.
+//
+// The caller is responsible for ensuring that the template Entry struct
+// does not change during the the lifetime of the returned *log.Logger.
+//
+// Prefer (*Logger).StandardLogger() which is more efficient if the template
+// only sets Severity.
+func (l *Logger) StandardLoggerFromTemplate(template *Entry) *log.Logger {
+	return log.New(templateEntryWriter{l, template}, "", 0)
+}
+
+func populateTraceInfo(e *Entry, req *http.Request) bool {
+	if req == nil {
+		if e.HTTPRequest != nil && e.HTTPRequest.Request != nil {
+			req = e.HTTPRequest.Request
+		} else {
+			return false
+		}
+	}
+	otelSpanContext := trace.SpanContextFromContext(req.Context())
+	if otelSpanContext.IsValid() {
+		e.Trace = otelSpanContext.TraceID().String()
+		e.SpanID = otelSpanContext.SpanID().String()
+		e.TraceSampled = otelSpanContext.IsSampled()
+		return true
+	}
+	header := req.Header.Get("Traceparent")
+	if header != "" {
+		// do not use traceSampled flag defined by traceparent because
+		// flag's definition differs from expected by Cloud Tracing
+		traceID, spanID, _ := deconstructTraceParent(header)
+		if traceID != "" {
+			e.Trace = traceID
+			e.SpanID = spanID
+			return true
+		}
+	}
+	header = req.Header.Get("X-Cloud-Trace-Context")
+	if header != "" {
+		traceID, spanID, traceSampled := deconstructXCloudTraceContext(header)
+		if traceID != "" {
+			e.Trace = traceID
+			e.SpanID = spanID
+			// enforce sampling if required
+			e.TraceSampled = e.TraceSampled || traceSampled
+			return true
+		}
+	}
+	return false
+}
+
+// As per format described at https://www.w3.org/TR/trace-context/#traceparent-header-field-values
+var validTraceParentExpression = regexp.MustCompile(`^(00)-([a-fA-F\d]{32})-([a-f\d]{16})-([a-fA-F\d]{2})$`)
+
+func deconstructTraceParent(s string) (traceID, spanID string, traceSampled bool) {
+	matches := validTraceParentExpression.FindStringSubmatch(s)
+	if matches != nil {
+		// regexp package does not support negative lookahead preventing all 0 validations
+		if matches[2] == "00000000000000000000000000000000" || matches[3] == "0000000000000000" {
+			return
+		}
+		flags, err := strconv.ParseInt(matches[4], 16, 16)
+		if err == nil {
+			traceSampled = (flags & 0x01) == 1
+		}
+		traceID, spanID = matches[2], matches[3]
+	}
+	return
+}
+
+var validXCloudTraceContext = regexp.MustCompile(
+	// Matches on "TRACE_ID"
+	`([a-f\d]+)?` +
+		// Matches on "/SPAN_ID"
+		`(?:/([a-f\d]+))?` +
+		// Matches on ";0=TRACE_TRUE"
+		`(?:;o=(\d))?`)
 
 func deconstructXCloudTraceContext(s string) (traceID, spanID string, traceSampled bool) {
-	// As per the format described at https://cloud.google.com/trace/docs/troubleshooting#force-trace
+	// As per the format described at https://cloud.google.com/trace/docs/setup#force-trace
 	//    "X-Cloud-Trace-Context: TRACE_ID/SPAN_ID;o=TRACE_TRUE"
 	// for example:
-	//    "X-Cloud-Trace-Context: 105445aa7843bc8bf206b120001000/0;o=1"
+	//    "X-Cloud-Trace-Context: 105445aa7843bc8bf206b120001000/1;o=1"
 	//
 	// We expect:
-	//   * traceID:         "105445aa7843bc8bf206b120001000"
-	//   * spanID:          ""
-	//   * traceSampled:    true
-	matches := reCloudTraceContext.FindAllStringSubmatch(s, -1)
-	if len(matches) != 1 {
-		return
+	//   * traceID (optional): 			"105445aa7843bc8bf206b120001000"
+	//   * spanID (optional):       	"1"
+	//   * traceSampled (optional): 	true
+	matches := validXCloudTraceContext.FindStringSubmatch(s)
+
+	if matches != nil {
+		traceID, spanID, traceSampled = matches[1], matches[2], matches[3] == "1"
 	}
 
-	sub := matches[0]
-	if len(sub) != 4 {
-		return
-	}
-
-	traceID, spanID = sub[1], sub[2]
 	if spanID == "0" {
 		spanID = ""
 	}
-	traceSampled = sub[3] == "1"
 
 	return
 }
 
-func (l *Logger) toLogEntry(e Entry) (*logpb.LogEntry, error) {
+// ToLogEntry takes an Entry structure and converts it to the LogEntry proto.
+// A parent can take any of the following forms:
+//
+//	projects/PROJECT_ID
+//	folders/FOLDER_ID
+//	billingAccounts/ACCOUNT_ID
+//	organizations/ORG_ID
+//
+// for backwards compatibility, a string with no '/' is also allowed and is interpreted
+// as a project ID.
+//
+// ToLogEntry is implied when users invoke Logger.Log or Logger.LogSync,
+// but its exported as a pub function here to give users additional flexibility
+// when using the library. Don't call this method manually if Logger.Log or
+// Logger.LogSync are used, it is intended to be used together with direct call
+// to WriteLogEntries method.
+func ToLogEntry(e Entry, parent string) (*logpb.LogEntry, error) {
+	var l Logger
+	return l.ToLogEntry(e, parent)
+}
+
+// ToLogEntry for Logger instance
+func (l *Logger) ToLogEntry(e Entry, parent string) (*logpb.LogEntry, error) {
+	parent, err := makeParent(parent)
+	if err != nil {
+		return nil, err
+	}
+	return toLogEntryInternal(e, l, parent, 1)
+}
+
+func toLogEntryInternalImpl(e Entry, l *Logger, parent string, skipLevels int) (*logpb.LogEntry, error) {
 	if e.LogName != "" {
 		return nil, errors.New("logging: Entry.LogName should be not be set when writing")
 	}
@@ -892,34 +935,43 @@ func (l *Logger) toLogEntry(e Entry) (*logpb.LogEntry, error) {
 	if t.IsZero() {
 		t = now()
 	}
-	ts, err := ptypes.TimestampProto(t)
-	if err != nil {
-		return nil, err
+	ts := timestamppb.New(t)
+	if l != nil && l.populateSourceLocation != DoNotPopulateSourceLocation && e.SourceLocation == nil {
+		if l.populateSourceLocation == AlwaysPopulateSourceLocation ||
+			l.populateSourceLocation == PopulateSourceLocationForDebugEntries && e.Severity == Severity(Debug) {
+			// filename and line are captured for source code that calls
+			// skipLevels up the goroutine calling stack + 1 for this func.
+			pc, file, line, ok := runtime.Caller(skipLevels + 1)
+			if ok {
+				details := runtime.FuncForPC(pc)
+				e.SourceLocation = &logpb.LogEntrySourceLocation{
+					File:     file,
+					Function: details.Name(),
+					Line:     int64(line),
+				}
+			}
+		}
 	}
-	if e.Trace == "" && e.HTTPRequest != nil && e.HTTPRequest.Request != nil {
-		traceHeader := e.HTTPRequest.Request.Header.Get("X-Cloud-Trace-Context")
-		if traceHeader != "" {
-			// Set to a relative resource name, as described at
-			// https://cloud.google.com/appengine/docs/flexible/go/writing-application-logs.
-			traceID, spanID, traceSampled := deconstructXCloudTraceContext(traceHeader)
-			if traceID != "" {
-				e.Trace = fmt.Sprintf("%s/traces/%s", l.client.parent, traceID)
-			}
-			if e.SpanID == "" {
-				e.SpanID = spanID
-			}
-
-			// If we previously hadn't set TraceSampled, let's retrieve it
-			// from the HTTP request's header, as per:
-			//   https://cloud.google.com/trace/docs/troubleshooting#force-trace
-			e.TraceSampled = e.TraceSampled || traceSampled
+	if e.Trace == "" {
+		populateTraceInfo(&e, nil)
+		// format trace
+		if e.Trace != "" && !strings.Contains(e.Trace, "/traces/") {
+			e.Trace = fmt.Sprintf("%s/traces/%s", parent, e.Trace)
+		}
+	}
+	req, err := fromHTTPRequest(e.HTTPRequest)
+	if err != nil {
+		if l != nil && l.client != nil {
+			l.client.error(err)
+		} else {
+			return nil, err
 		}
 	}
 	ent := &logpb.LogEntry{
 		Timestamp:      ts,
 		Severity:       logtypepb.LogSeverity(e.Severity),
 		InsertId:       e.InsertID,
-		HttpRequest:    fromHTTPRequest(e.HTTPRequest),
+		HttpRequest:    req,
 		Operation:      e.Operation,
 		Labels:         e.Labels,
 		Trace:          e.Trace,
@@ -931,6 +983,8 @@ func (l *Logger) toLogEntry(e Entry) (*logpb.LogEntry, error) {
 	switch p := e.Payload.(type) {
 	case string:
 		ent.Payload = &logpb.LogEntry_TextPayload{TextPayload: p}
+	case *anypb.Any:
+		ent.Payload = &logpb.LogEntry_ProtoPayload{ProtoPayload: p}
 	default:
 		s, err := toProtoStruct(p)
 		if err != nil {
@@ -940,3 +994,88 @@ func (l *Logger) toLogEntry(e Entry) (*logpb.LogEntry, error) {
 	}
 	return ent, nil
 }
+
+// entry represents the fields of a logging.Entry that can be parsed by Logging agent.
+// See the mappings at https://cloud.google.com/logging/docs/structured-logging#special-payload-fields
+type structuredLogEntry struct {
+	Message        json.RawMessage                   `json:"message"`
+	Severity       string                            `json:"severity,omitempty"`
+	HTTPRequest    *structuredLogEntryHTTPRequest    `json:"httpRequest,omitempty"`
+	Timestamp      string                            `json:"timestamp,omitempty"`
+	Labels         map[string]string                 `json:"logging.googleapis.com/labels,omitempty"`
+	InsertID       string                            `json:"logging.googleapis.com/insertId,omitempty"`
+	Operation      *structuredLogEntryOperation      `json:"logging.googleapis.com/operation,omitempty"`
+	SourceLocation *structuredLogEntrySourceLocation `json:"logging.googleapis.com/sourceLocation,omitempty"`
+	SpanID         string                            `json:"logging.googleapis.com/spanId,omitempty"`
+	Trace          string                            `json:"logging.googleapis.com/trace,omitempty"`
+	TraceSampled   bool                              `json:"logging.googleapis.com/trace_sampled,omitempty"`
+}
+
+// structuredLogEntryHTTPRequest wraps the HTTPRequest proto field in structuredLogEntry for easier JSON marshalling.
+type structuredLogEntryHTTPRequest struct {
+	request *logtypepb.HttpRequest
+}
+
+func (s structuredLogEntryHTTPRequest) MarshalJSON() ([]byte, error) {
+	return protojson.Marshal(s.request)
+}
+
+// structuredLogEntryOperation wraps the Operation proto field in structuredLogEntry for easier JSON marshalling.
+type structuredLogEntryOperation struct {
+	operation *logpb.LogEntryOperation
+}
+
+func (s structuredLogEntryOperation) MarshalJSON() ([]byte, error) {
+	return protojson.Marshal(s.operation)
+}
+
+// structuredLogEntrySourceLocation wraps the SourceLocation proto field in structuredLogEntry for easier JSON marshalling.
+type structuredLogEntrySourceLocation struct {
+	sourceLocation *logpb.LogEntrySourceLocation
+}
+
+func (s structuredLogEntrySourceLocation) MarshalJSON() ([]byte, error) {
+	return protojson.Marshal(s.sourceLocation)
+}
+
+func serializeEntryToWriter(entry *logpb.LogEntry, w io.Writer) error {
+	var httpRequest *structuredLogEntryHTTPRequest
+	if entry.HttpRequest != nil {
+		httpRequest = &structuredLogEntryHTTPRequest{entry.HttpRequest}
+	}
+
+	var operation *structuredLogEntryOperation
+	if entry.Operation != nil {
+		operation = &structuredLogEntryOperation{entry.Operation}
+	}
+
+	var sourceLocation *structuredLogEntrySourceLocation
+	if entry.SourceLocation != nil {
+		sourceLocation = &structuredLogEntrySourceLocation{entry.SourceLocation}
+	}
+
+	jsonifiedEntry := structuredLogEntry{
+		Severity:       entry.Severity.String(),
+		HTTPRequest:    httpRequest,
+		Timestamp:      entry.Timestamp.String(),
+		Labels:         entry.Labels,
+		InsertID:       entry.InsertId,
+		Operation:      operation,
+		SourceLocation: sourceLocation,
+		SpanID:         entry.SpanId,
+		Trace:          entry.Trace,
+		TraceSampled:   entry.TraceSampled,
+	}
+	var err error
+	if entry.GetTextPayload() != "" {
+		jsonifiedEntry.Message, err = json.Marshal(entry.GetTextPayload())
+	} else if entry.GetJsonPayload() != nil {
+		jsonifiedEntry.Message, err = json.Marshal(entry.GetJsonPayload().AsMap())
+	} else {
+		return ErrRedirectProtoPayloadNotSupported
+	}
+	if err == nil {
+		err = json.NewEncoder(w).Encode(jsonifiedEntry)
+	}
+	return err
+}
