@@ -25,23 +25,31 @@ import (
 	"golang.org/x/sync/errgroup"
 )
 
+func fcSettings(c int, s int, l LimitExceededBehavior) FlowControlSettings {
+	return FlowControlSettings{
+		MaxOutstandingMessages: c,
+		MaxOutstandingBytes:    s,
+		LimitExceededBehavior:  l,
+	}
+}
+
 func TestFlowControllerCancel(t *testing.T) {
 	// Test canceling a flow controller's context.
 	t.Parallel()
-	fc := newFlowController(3, 10)
+	fc := newFlowController(fcSettings(3, 10, FlowControlBlock))
 	if err := fc.acquire(context.Background(), 5); err != nil {
 		t.Fatal(err)
 	}
 	// Experiment: a context that times out should always return an error.
 	ctx, cancel := context.WithTimeout(context.Background(), 5*time.Millisecond)
 	defer cancel()
-	if err := fc.acquire(ctx, 6); err != context.DeadlineExceeded {
+	if err := fc.acquire(ctx, 6); !errors.Is(err, context.DeadlineExceeded) {
 		t.Fatalf("got %v, expected DeadlineExceeded", err)
 	}
 	// Control: a context that is not done should always return nil.
 	go func() {
 		time.Sleep(5 * time.Millisecond)
-		fc.release(5)
+		fc.release(ctx, 5)
 	}()
 	if err := fc.acquire(context.Background(), 6); err != nil {
 		t.Errorf("got %v, expected nil", err)
@@ -51,7 +59,7 @@ func TestFlowControllerCancel(t *testing.T) {
 func TestFlowControllerLargeRequest(t *testing.T) {
 	// Large requests succeed, consuming the entire allotment.
 	t.Parallel()
-	fc := newFlowController(3, 10)
+	fc := newFlowController(fcSettings(3, 10, FlowControlBlock))
 	err := fc.acquire(context.Background(), 11)
 	if err != nil {
 		t.Fatal(err)
@@ -64,13 +72,13 @@ func TestFlowControllerNoStarve(t *testing.T) {
 	t.Parallel()
 	ctx, cancel := context.WithTimeout(context.Background(), 2*time.Second)
 	defer cancel()
-	fc := newFlowController(10, 10)
+	fc := newFlowController(fcSettings(10, 10, FlowControlBlock))
 	first := make(chan int)
 	for i := 0; i < 20; i++ {
 		go func() {
 			for {
 				if err := fc.acquire(ctx, 1); err != nil {
-					if err != context.Canceled {
+					if !errors.Is(err, context.Canceled) {
 						t.Error(err)
 					}
 					return
@@ -79,7 +87,7 @@ func TestFlowControllerNoStarve(t *testing.T) {
 				case first <- 1:
 				default:
 				}
-				fc.release(1)
+				fc.release(ctx, 1)
 			}
 		}()
 	}
@@ -120,7 +128,7 @@ func TestFlowControllerSaturation(t *testing.T) {
 			wantSize:    9,
 		},
 	} {
-		fc := newFlowController(maxCount, maxSize)
+		fc := newFlowController(fcSettings(maxCount, maxSize, FlowControlBlock))
 		// Atomically track flow controller state.
 		// The flowController itself tracks count.
 		var curSize int64
@@ -162,7 +170,7 @@ func TestFlowControllerSaturation(t *testing.T) {
 					if atomic.AddInt64(&curSize, -int64(test.acquireSize)) < 0 {
 						return errors.New("negative size")
 					}
-					fc.release(test.acquireSize)
+					fc.release(ctx, test.acquireSize)
 				}
 				return success
 			})
@@ -174,59 +182,39 @@ func TestFlowControllerSaturation(t *testing.T) {
 	}
 }
 
-func TestFlowControllerTryAcquire(t *testing.T) {
-	t.Parallel()
-	fc := newFlowController(3, 10)
-
-	// Successfully tryAcquire 4 bytes.
-	if !fc.tryAcquire(4) {
-		t.Error("got false, wanted true")
-	}
-
-	// Fail to tryAcquire 7 bytes.
-	if fc.tryAcquire(7) {
-		t.Error("got true, wanted false")
-	}
-
-	// Successfully tryAcquire 6 byte.
-	if !fc.tryAcquire(6) {
-		t.Error("got false, wanted true")
-	}
-}
-
 func TestFlowControllerUnboundedCount(t *testing.T) {
 	t.Parallel()
 	ctx := context.Background()
-	fc := newFlowController(0, 10)
+	fc := newFlowController(fcSettings(0, 10, FlowControlSignalError))
 
 	// Successfully acquire 4 bytes.
 	if err := fc.acquire(ctx, 4); err != nil {
 		t.Errorf("got %v, wanted no error", err)
 	}
 
-	// Successfully tryAcquire 4 bytes.
-	if !fc.tryAcquire(4) {
-		t.Error("got false, wanted true")
+	// Successfully acquire 4 bytes.
+	if err := fc.acquire(ctx, 4); err != nil {
+		t.Errorf("got %v, wanted no error", err)
 	}
 
-	// Fail to tryAcquire 3 bytes.
-	if fc.tryAcquire(3) {
-		t.Error("got true, wanted false")
+	// Fail to acquire 3 bytes.
+	if err := fc.acquire(ctx, 3); err == nil {
+		t.Errorf("got nil, wanted %v", ErrFlowControllerMaxOutstandingBytes)
 	}
 }
 
 func TestFlowControllerUnboundedCount2(t *testing.T) {
 	t.Parallel()
 	ctx := context.Background()
-	fc := newFlowController(0, 0)
+	fc := newFlowController(fcSettings(0, 0, FlowControlSignalError))
 	// Successfully acquire 4 bytes.
 	if err := fc.acquire(ctx, 4); err != nil {
 		t.Errorf("got %v, wanted no error", err)
 	}
-	fc.release(1)
-	fc.release(1)
-	fc.release(1)
-	wantCount := int64(-2)
+	fc.release(ctx, 1)
+	fc.release(ctx, 1)
+	fc.release(ctx, 1)
+	wantCount := int64(0)
 	c := int64(fc.count())
 	if c != wantCount {
 		t.Fatalf("got count %d, want %d", c, wantCount)
@@ -236,20 +224,20 @@ func TestFlowControllerUnboundedCount2(t *testing.T) {
 func TestFlowControllerUnboundedBytes(t *testing.T) {
 	t.Parallel()
 	ctx := context.Background()
-	fc := newFlowController(2, 0)
+	fc := newFlowController(fcSettings(2, 0, FlowControlSignalError))
 
 	// Successfully acquire 4GB.
 	if err := fc.acquire(ctx, 4e9); err != nil {
 		t.Errorf("got %v, wanted no error", err)
 	}
 
-	// Successfully tryAcquire 4GB bytes.
-	if !fc.tryAcquire(4e9) {
-		t.Error("got false, wanted true")
+	// Successfully acquired 4GB bytes.
+	if err := fc.acquire(ctx, 4e9); err != nil {
+		t.Errorf("got %v, wanted no error", err)
 	}
 
-	// Fail to tryAcquire a third message.
-	if fc.tryAcquire(3) {
-		t.Error("got true, wanted false")
+	// Fail to acquire a third message.
+	if err := fc.acquire(ctx, 3); err == nil {
+		t.Errorf("got nil, wanted %v", ErrFlowControllerMaxOutstandingMessages)
 	}
 }
