@@ -19,12 +19,14 @@ package spanner
 import (
 	"bytes"
 	"fmt"
+	"math/big"
 	"time"
 
 	"cloud.google.com/go/civil"
-	proto3 "github.com/golang/protobuf/ptypes/struct"
-	sppb "google.golang.org/genproto/googleapis/spanner/v1"
+	sppb "cloud.google.com/go/spanner/apiv1/spannerpb"
 	"google.golang.org/grpc/codes"
+	"google.golang.org/protobuf/reflect/protoreflect"
+	proto3 "google.golang.org/protobuf/types/known/structpb"
 )
 
 // A Key can be either a Cloud Spanner row's primary key or a secondary index
@@ -57,6 +59,7 @@ import (
 //   - string and NullString are mapped to Cloud Spanner's STRING type.
 //   - time.Time and NullTime are mapped to Cloud Spanner's TIMESTAMP type.
 //   - civil.Date and NullDate are mapped to Cloud Spanner's DATE type.
+//   - protoreflect.Enum and NullProtoEnum are mapped to Cloud Spanner's ENUM type.
 type Key []interface{}
 
 // errInvdKeyPartType returns error for unsupported key part type.
@@ -82,10 +85,14 @@ func keyPartValue(part interface{}) (pb *proto3.Value, err error) {
 		pb, _, err = encodeValue(int64(v))
 	case uint32:
 		pb, _, err = encodeValue(int64(v))
-	case float32:
-		pb, _, err = encodeValue(float64(v))
-	case int64, float64, NullInt64, NullFloat64, bool, NullBool, []byte, string, NullString, time.Time, civil.Date, NullTime, NullDate:
+	case int64, float64, float32, NullInt64, NullFloat64, NullFloat32, bool, NullBool, []byte, string, NullString, time.Time, civil.Date, NullTime, NullDate, big.Rat, NullNumeric, protoreflect.Enum, NullProtoEnum:
 		pb, _, err = encodeValue(v)
+	case Encoder:
+		part, err = v.EncodeSpanner()
+		if err != nil {
+			return nil, err
+		}
+		pb, err = keyPartValue(part)
 	default:
 		return nil, errInvdKeyPartType(v)
 	}
@@ -125,40 +132,54 @@ func (key Key) String() string {
 		if i != 0 {
 			fmt.Fprint(b, ",")
 		}
-		switch v := part.(type) {
-		case int, int8, int16, int32, int64, uint, uint8, uint16, uint32, float32, float64, bool:
-			// Use %v to print numeric types and bool.
-			fmt.Fprintf(b, "%v", v)
-		case string:
-			fmt.Fprintf(b, "%q", v)
-		case []byte:
-			if v != nil {
-				fmt.Fprintf(b, "%q", v)
-			} else {
-				fmt.Fprint(b, nullString)
-			}
-		case NullInt64, NullFloat64, NullBool:
-			// The above types implement fmt.Stringer.
-			fmt.Fprintf(b, "%s", v)
-		case NullString, NullDate, NullTime:
-			// Quote the returned string if it is not null.
-			if v.(NullableValue).IsNull() {
-				fmt.Fprintf(b, "%s", nullString)
-			} else {
-				fmt.Fprintf(b, "%q", v)
-			}
-		case civil.Date:
-			fmt.Fprintf(b, "%q", v)
-		case time.Time:
-			fmt.Fprintf(b, "%q", v.Format(time.RFC3339Nano))
-		default:
-			fmt.Fprintf(b, "%v", v)
-		}
+		key.elemString(b, part)
 	}
 	fmt.Fprint(b, ")")
 	return b.String()
 }
 
+func (key Key) elemString(b *bytes.Buffer, part interface{}) {
+	switch v := part.(type) {
+	case int, int8, int16, int32, int64, uint, uint8, uint16, uint32, float32, float64, bool, protoreflect.Enum:
+		// Use %v to print numeric types and bool.
+		fmt.Fprintf(b, "%v", v)
+	case string:
+		fmt.Fprintf(b, "%q", v)
+	case []byte:
+		if v != nil {
+			fmt.Fprintf(b, "%q", v)
+		} else {
+			fmt.Fprint(b, nullString)
+		}
+	case NullInt64, NullFloat64, NullBool, NullNumeric, NullProtoEnum:
+		// The above types implement fmt.Stringer.
+		fmt.Fprintf(b, "%s", v)
+	case NullString, NullDate, NullTime:
+		// Quote the returned string if it is not null.
+		if v.(NullableValue).IsNull() {
+			fmt.Fprintf(b, "%s", nullString)
+		} else {
+			fmt.Fprintf(b, "%q", v)
+		}
+	case civil.Date:
+		fmt.Fprintf(b, "%q", v)
+	case time.Time:
+		fmt.Fprintf(b, "%q", v.Format(time.RFC3339Nano))
+	case big.Rat:
+		fmt.Fprintf(b, "%v", NumericString(&v))
+	case Encoder:
+		var err error
+		part, err = v.EncodeSpanner()
+		if err != nil {
+			fmt.Fprintf(b, "error")
+		} else {
+			key.elemString(b, part)
+		}
+	default:
+		fmt.Fprintf(b, "%v", v)
+	}
+}
+
 // AsPrefix returns a KeyRange for all keys where k is the prefix.
 func (key Key) AsPrefix() KeyRange {
 	return KeyRange{
@@ -214,7 +235,7 @@ const (
 // the following range returns all events for user "Bob" that occurred in the
 // year 2015:
 //
-// 	spanner.KeyRange{
+//	spanner.KeyRange{
 //		Start: spanner.Key{"Bob", "2015-01-01"},
 //		End:   spanner.Key{"Bob", "2015-12-31"},
 //		Kind:  ClosedClosed,
@@ -386,6 +407,15 @@ func KeySets(keySets ...KeySet) KeySet {
 	return u
 }
 
+// KeySetFromKeys returns a KeySet containing the given slice of keys.
+func KeySetFromKeys(keys ...Key) KeySet {
+	u := make(union, len(keys))
+	for i, k := range keys {
+		u[i] = k
+	}
+	return u
+}
+
 type union []KeySet
 
 func (u union) keySetProto() (*sppb.KeySet, error) {
