@@ -17,12 +17,15 @@ limitations under the License.
 package spanner
 
 import (
+	"math/big"
 	"sort"
 	"strings"
 	"testing"
+	"time"
 
-	proto3 "github.com/golang/protobuf/ptypes/struct"
-	sppb "google.golang.org/genproto/googleapis/spanner/v1"
+	"cloud.google.com/go/civil"
+	sppb "cloud.google.com/go/spanner/apiv1/spannerpb"
+	proto3 "google.golang.org/protobuf/types/known/structpb"
 )
 
 // keysetProto returns protobuf encoding of valid spanner.KeySet.
@@ -36,6 +39,12 @@ func keysetProto(t *testing.T, ks KeySet) *sppb.KeySet {
 
 // Test encoding from spanner.Mutation to protobuf.
 func TestMutationToProto(t *testing.T) {
+	utc, err := time.LoadLocation("UTC")
+	if err != nil {
+		t.Fatalf("Could not load UTC: %v", err)
+	}
+	r, _ := (&big.Rat{}).SetString("3.14")
+
 	for i, test := range []struct {
 		m    *Mutation
 		want *sppb.Mutation
@@ -120,9 +129,50 @@ func TestMutationToProto(t *testing.T) {
 				},
 			},
 		},
+		// Mutation with all supported data types
+		{
+			&Mutation{
+				opInsert,
+				"t_foo",
+				KeySets(),
+				[]string{"colBool", "colInt64", "colFloat64", "colNumeric", "colString", "colBytes", "colDate", "colTimestamp"},
+				[]interface{}{
+					true,
+					int64(100),
+					float64(3.14),
+					*r, // 3.14
+					"one",
+					[]byte{1, 2, 3},
+					civil.Date{Year: 2020, Month: 12, Day: 2},
+					time.Date(2020, time.December, 3, 8, 46, 58, 109, utc),
+				},
+			},
+			&sppb.Mutation{
+				Operation: &sppb.Mutation_Insert{
+					Insert: &sppb.Mutation_Write{
+						Table:   "t_foo",
+						Columns: []string{"colBool", "colInt64", "colFloat64", "colNumeric", "colString", "colBytes", "colDate", "colTimestamp"},
+						Values: []*proto3.ListValue{
+							{
+								Values: []*proto3.Value{
+									boolProto(true),
+									stringProto("100"),
+									floatProto(3.14),
+									stringProto("3.140000000"),
+									stringProto("one"),
+									bytesProto([]byte{1, 2, 3}),
+									stringProto("2020-12-02"),
+									stringProto("2020-12-03T08:46:58.000000109Z"),
+								},
+							},
+						},
+					},
+				},
+			},
+		},
 	} {
 		if got, err := test.m.proto(); err != nil || !testEqual(got, test.want) {
-			t.Errorf("%d: (%#v).proto() = (%v, %v), want (%v, nil)", i, test.m, got, err, test.want)
+			t.Errorf("%d:\n(%#v).proto() =\n     (%v, %v)\nwant (%v, nil)", i, test.m, got, err, test.want)
 		}
 	}
 }
@@ -568,3 +618,110 @@ func TestEncodeMutationArray(t *testing.T) {
 		}
 	}
 }
+
+func TestEncodeMutationGroupArray(t *testing.T) {
+	for _, test := range []struct {
+		name    string
+		mgs     []*MutationGroup
+		want    []*sppb.BatchWriteRequest_MutationGroup
+		wantErr error
+	}{
+		{
+			"Multiple Mutations",
+			[]*MutationGroup{
+				{[]*Mutation{
+					{opDelete, "t_test", Key{"bar"}, nil, nil},
+					{opInsertOrUpdate, "t_test", nil, []string{"key", "val"}, []interface{}{"foo1", 1}},
+				}},
+				{[]*Mutation{
+					{opInsert, "t_test", nil, []string{"key", "val"}, []interface{}{"foo2", 1}},
+					{opUpdate, "t_test", nil, []string{"key", "val"}, []interface{}{"foo3", 1}},
+				}},
+				{[]*Mutation{
+					{opReplace, "t_test", nil, []string{"key", "val"}, []interface{}{"foo4", 1}},
+				}},
+			},
+			[]*sppb.BatchWriteRequest_MutationGroup{
+				{Mutations: []*sppb.Mutation{
+					{
+						Operation: &sppb.Mutation_Delete_{
+							Delete: &sppb.Mutation_Delete{
+								Table: "t_test",
+								KeySet: &sppb.KeySet{
+									Keys: []*proto3.ListValue{listValueProto(stringProto("bar"))},
+								},
+							},
+						},
+					},
+					{
+						Operation: &sppb.Mutation_InsertOrUpdate{
+							InsertOrUpdate: &sppb.Mutation_Write{
+								Table:   "t_test",
+								Columns: []string{"key", "val"},
+								Values:  []*proto3.ListValue{listValueProto(stringProto("foo1"), intProto(1))},
+							},
+						},
+					},
+				}},
+				{Mutations: []*sppb.Mutation{
+					{
+						Operation: &sppb.Mutation_Insert{
+							Insert: &sppb.Mutation_Write{
+								Table:   "t_test",
+								Columns: []string{"key", "val"},
+								Values:  []*proto3.ListValue{listValueProto(stringProto("foo2"), intProto(1))},
+							},
+						},
+					},
+					{
+						Operation: &sppb.Mutation_Update{
+							Update: &sppb.Mutation_Write{
+								Table:   "t_test",
+								Columns: []string{"key", "val"},
+								Values:  []*proto3.ListValue{listValueProto(stringProto("foo3"), intProto(1))},
+							},
+						},
+					},
+				}},
+				{Mutations: []*sppb.Mutation{
+					{
+						Operation: &sppb.Mutation_Replace{
+							Replace: &sppb.Mutation_Write{
+								Table:   "t_test",
+								Columns: []string{"key", "val"},
+								Values:  []*proto3.ListValue{listValueProto(stringProto("foo4"), intProto(1))},
+							},
+						},
+					},
+				}},
+			},
+			nil,
+		},
+		{
+			"Multiple Mutations - Bad Mutation",
+			[]*MutationGroup{
+				{[]*Mutation{
+					{opDelete, "t_test", Key{"bar"}, nil, nil},
+					{opInsertOrUpdate, "t_test", nil, []string{"key", "val"}, []interface{}{"foo1", struct{}{}}},
+				}},
+				{[]*Mutation{
+					{opInsert, "t_test", nil, []string{"key", "val"}, []interface{}{"foo2", 1}},
+					{opUpdate, "t_test", nil, []string{"key", "val"}, []interface{}{"foo3", 1}},
+				}},
+			},
+			[]*sppb.BatchWriteRequest_MutationGroup{},
+			errEncoderUnsupportedType(struct{}{}),
+		},
+	} {
+		gotProto, gotErr := mutationGroupsProto(test.mgs)
+		if gotErr != nil {
+			if !testEqual(gotErr, test.wantErr) {
+				t.Errorf("%v: mutationGroupsProto(%v) returns error %v, want %v", test.name, test.mgs, gotErr, test.wantErr)
+			}
+			continue
+		}
+		if !testEqual(gotProto, test.want) {
+			t.Errorf("%v: mutationGroupsProto(%v) = (%v, nil), want (%v, nil)", test.name, test.mgs, gotProto, test.want)
+		}
+	}
+}
