@@ -17,14 +17,16 @@ limitations under the License.
 package spanner
 
 import (
+	"bytes"
 	"math"
 	"testing"
 	"time"
 
 	"cloud.google.com/go/civil"
-	"github.com/golang/protobuf/proto"
-	proto3 "github.com/golang/protobuf/ptypes/struct"
-	sppb "google.golang.org/genproto/googleapis/spanner/v1"
+	sppb "cloud.google.com/go/spanner/apiv1/spannerpb"
+	"google.golang.org/protobuf/encoding/prototext"
+	"google.golang.org/protobuf/proto"
+	proto3 "google.golang.org/protobuf/types/known/structpb"
 )
 
 func TestConvertParams(t *testing.T) {
@@ -46,10 +48,15 @@ func TestConvertParams(t *testing.T) {
 	type staticStruct struct {
 		Field int `spanner:"field"`
 	}
+	type CustomInt int64
+	type staticStructWithCustomType struct {
+		Field CustomInt `spanner:"field"`
+	}
 
 	var (
 		s1 = staticStruct{10}
 		s2 = staticStruct{20}
+		s3 = staticStructWithCustomType{30}
 	)
 
 	for _, test := range []struct {
@@ -95,6 +102,19 @@ func TestConvertParams(t *testing.T) {
 		{[]NullFloat64(nil), nullProto(), listType(floatType())},
 		{[]NullFloat64{}, listProto(), listType(floatType())},
 		{[]NullFloat64{{2.72, true}, {}}, listProto(floatProto(2.72), nullProto()), listType(floatType())},
+		// float32
+		{float32(0.0), float32Proto(0.0), float32Type()},
+		{float32(math.Inf(1)), float32Proto(float32(math.Inf(1))), float32Type()},
+		{float32(math.Inf(-1)), float32Proto(float32(math.Inf(-1))), float32Type()},
+		{float32(math.NaN()), float32Proto(float32(math.NaN())), float32Type()},
+		{NullFloat32{3.14, true}, float32Proto(3.14), float32Type()},
+		{NullFloat32{-99.99, false}, nullProto(), float32Type()},
+		{[]float32(nil), nullProto(), listType(float32Type())},
+		{[]float32{}, listProto(), listType(float32Type())},
+		{[]float32{3.14, float32(math.Inf(1))}, listProto(float32Proto(3.14), float32Proto(float32(math.Inf(1)))), listType(float32Type())},
+		{[]NullFloat32(nil), nullProto(), listType(float32Type())},
+		{[]NullFloat32{}, listProto(), listType(float32Type())},
+		{[]NullFloat32{{3.14, true}, {}}, listProto(float32Proto(3.14), nullProto()), listType(float32Type())},
 		// string
 		{"", stringProto(""), stringType()},
 		{"foo", stringProto("foo"), stringType()},
@@ -133,6 +153,11 @@ func TestConvertParams(t *testing.T) {
 			listProto(intProto(10)),
 			structType(mkField("field", intType())),
 		},
+		{
+			s3,
+			listProto(intProto(30)),
+			structType(mkField("field", intType())),
+		},
 		{
 			(*struct {
 				F1 civil.Date `spanner:""`
@@ -149,6 +174,12 @@ func TestConvertParams(t *testing.T) {
 			listProto(listProto(intProto(10)), listProto(intProto(20))),
 			listType(structType(mkField("field", intType()))),
 		},
+		// Untyped null
+		{
+			nil,
+			nullProto(),
+			nil,
+		},
 	} {
 		st.Params["var"] = test.val
 		gotParams, gotParamTypes, gotErr := st.convertParams()
@@ -159,31 +190,25 @@ func TestConvertParams(t *testing.T) {
 		gotParamField := gotParams.Fields["var"]
 		if !proto.Equal(gotParamField, test.wantField) {
 			// handle NaN
-			if test.wantType.Code == floatType().Code && proto.MarshalTextString(gotParamField) == proto.MarshalTextString(test.wantField) {
+			gotParamFieldText, err := prototext.Marshal(gotParamField)
+			if err != nil {
+				t.Fatal(err)
+			}
+			wantParamFieldText, err := prototext.Marshal(test.wantField)
+			if err != nil {
+				t.Fatal(err)
+			}
+			if test.wantType.Code == floatType().Code && bytes.Equal(gotParamFieldText, wantParamFieldText) {
 				continue
 			}
-			t.Errorf("%#v: got %v, want %v\n", test.val, gotParamField, test.wantField)
+			if test.wantType.Code == float32Type().Code && bytes.Equal(gotParamFieldText, wantParamFieldText) {
+				continue
+			}
+			t.Errorf("%#v:\n got: %v\nwant: %v\n", test.val, gotParamField, test.wantField)
 		}
 		gotParamType := gotParamTypes["var"]
 		if !proto.Equal(gotParamType, test.wantType) {
-			t.Errorf("%#v: got %v, want %v\n", test.val, gotParamType, test.wantField)
-		}
-	}
-
-	// Verify type error reporting.
-	for _, test := range []struct {
-		val     interface{}
-		wantErr error
-	}{
-		{
-			nil,
-			errBindParam("var", nil, errNilParam),
-		},
-	} {
-		st.Params["var"] = test.val
-		_, _, gotErr := st.convertParams()
-		if !testEqual(gotErr, test.wantErr) {
-			t.Errorf("value %#v:\ngot:  %v\nwant: %v", test.val, gotErr, test.wantErr)
+			t.Errorf("%#v:\n got: %v\nwant: %v\n", test.val, gotParamType, test.wantField)
 		}
 	}
 }
